#
#   Authors:
#       Pavel Brezina <pbrezina@redhat.com>
#
#   Copyright (C) 2017 Red Hat
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from collections import OrderedDict
from sbus_Invoker import Invoker, InvokerArgumentType, InvokerCaller, InvokerKeygen
from sbus_Introspection import SBus
from sbus_DataType import DataType


class Generator:
    @staticmethod
    def GenerateCode(templates, interfaces):
        """
            Generate asynchronous code for given interfaces.
        """

        class Callers:
            def __init__(self, interfaces, type):
                self.Methods = InvokerCaller.GatherMethodInvokers(interfaces,
                                                                  type)
                self.Signals = InvokerCaller.GatherSignalInvokers(interfaces,
                                                                  type)
                self.Getters = InvokerCaller.GatherGetInvokers(interfaces,
                                                               type)
                self.Setters = InvokerCaller.GatherSetInvokers(interfaces,
                                                               type)

        invokers = Invoker.GatherInvokers(interfaces)
        arguments = InvokerArgumentType.GatherArgumentTypes(interfaces)
        keygens = InvokerKeygen.GatherKeygens(interfaces)
        sync_callers = Callers(interfaces, "sync")
        async_callers = Callers(interfaces, "async")

        generators = [
            Generator.Interfaces(templates.get("interface.h"),
                                 interfaces),

            Generator.Symbols(templates.get("symbols.c"),
                              templates.get("symbols.h"),
                              interfaces),

            Generator.Arguments(templates.get("arguments.c"),
                                templates.get("arguments.h"),
                                arguments),

            Generator.Invokers(templates.get("invokers.c"),
                               templates.get("invokers.h"),
                               invokers),

            Generator.Keygens(templates.get("keygens.c"),
                              templates.get("keygens.h"),
                              keygens),

            Generator.Properties(templates.get("client_properties.h"),
                                 interfaces),

            Generator.MethodCalls(templates.get("client_async.c"),
                                  templates.get("client_async.h"),
                                  interfaces, "async", async_callers.Methods),

            Generator.SignalCalls(templates.get("client_async.c"),
                                  templates.get("client_async.h"),
                                  interfaces, "async", async_callers.Signals),

            Generator.PropertyCalls(templates.get("client_async.c"),
                                    templates.get("client_async.h"),
                                    interfaces, "async",
                                    async_callers.Getters,
                                    async_callers.Setters),

            Generator.MethodCalls(templates.get("client_sync.c"),
                                  templates.get("client_sync.h"),
                                  interfaces, "sync", sync_callers.Methods),

            Generator.SignalCalls(templates.get("client_sync.c"),
                                  templates.get("client_sync.h"),
                                  interfaces, "sync", sync_callers.Signals),

            Generator.PropertyCalls(templates.get("client_sync.c"),
                                    templates.get("client_sync.h"),
                                    interfaces, "sync",
                                    sync_callers.Getters,
                                    sync_callers.Setters)
        ]

        for generator in generators:
            generator.generate()

        templates.write()

    @staticmethod
    def FilterAnnotations(annotations):
        dict = OrderedDict()
        if annotations is None or not annotations:
            return dict

        for name, annotation in annotations.items():
            if not name.startswith("codegen."):
                dict[name] = annotation

        return dict

    class Base(object):
        """
            Base object for code generators.

            Children must implement generate() method.
        """

        def __init__(self):
            pass

        def generate(self):
            """ Make sure generate() method is implemented by children. """
            raise NotImplementedError("Method generate() is not implemented!")

        def tokenizeValue(self, *values):
            """ Concatenate valus into C token. """
            name = ""
            for value in values:
                name += '.' + value

            return name.strip('.').replace('.', '_')

        def tokenizeName(self, *values):
            """ Concatenate value.name into C token. """
            name = ""
            for value in values:
                name += '.' + value.name

            return name.strip('.').replace('.', '_')

        def setInputArguments(self, tpl, sbus_signature):
            """
                Set input arguments in template.
            """
            if Invoker.IsCustomInputHandler(sbus_signature):
                if tpl.hasLoop("in-raw"):
                    tpl.add("in-raw", {'type': "DBusMessageIter *"})
                return

            for idx, arg in enumerate(sbus_signature.arguments.values()):
                tpl.add("in", {'type': DataType.Find(arg.signature).inputCType,
                               'name': arg.name,
                               'index': idx})

        def setOutputArguments(self, tpl, sbus_signature):
            """
                Set output arguments in template.
            """
            if Invoker.IsCustomOutputHandler(sbus_signature):
                if tpl.hasLoop("out-raw"):
                    tpl.add("out-raw", {'type': "DBusMessageIter *"})
                return

            if tpl.hasLoop("out-static"):
                for idx, arg in enumerate(sbus_signature.arguments.values()):
                    type = DataType.Find(arg.signature)
                    if type.RequireTalloc:
                        continue

                    tpl.add("out-static", {'type': type.outputCType,
                                           'name': arg.name,
                                           'index': idx})

            if tpl.hasLoop("out-talloc"):
                for idx, arg in enumerate(sbus_signature.arguments.values()):
                    type = DataType.Find(arg.signature)
                    if not type.RequireTalloc:
                        continue

                    tpl.add("out-talloc", {'type': type.outputCType,
                                           'name': arg.name,
                                           'index': idx})

            if tpl.hasLoop("out"):
                for idx, arg in enumerate(sbus_signature.arguments.values()):
                    tpl.add("out", {
                        'type': DataType.Find(arg.signature).outputCType,
                        'name': arg.name,
                        'index': idx
                    })

        def getInterfaceName(self, iface):
            """
                If a codegen.Name annotation is specified, this name is used
                returned of a full D-Bus name.
            """
            annotation = "codegen.Name"
            iface_name = SBus.Annotation.Find(iface.annotations, annotation,
                                              iface.name)

            return self.tokenizeValue(iface_name)

        def getMemberName(self, iface, member):
            """
                If a codegen.Name annotation is specified, this name is used
                returned of a full D-Bus name.
            """
            annotation = "codegen.Name"
            member_name = SBus.Annotation.Find(member.annotations, annotation,
                                               member.name)

            return self.tokenizeValue(self.getInterfaceName(iface),
                                      member_name)

        def hasEmpty(self, invoker_signature):
            """
                Return true if the invoker signature has empty argument list.
            """
            return invoker_signature.invokerSignature == ""

        def hasRaw(self, invoker_signature):
            """
                Return true if the invoker signature has a custom handler.
            """
            return invoker_signature.invokerSignature == "raw"

        def hasAny(self, invoker_signature):
            """
                Return true if the invoker signature has any arguments.
            """
            return bool(invoker_signature.arguments) or \
                self.hasRaw(invoker_signature)

        def hasParsable(self, invoker_signature):
            """
                Return true if the invoker signature has any parsable
                arguments.
            """
            return invoker_signature.invokerSignature not in ["", "raw"]

        def outputNeedTalloc(self, invoker_signature):
            if self.hasRaw(invoker_signature):
                return True

            for idx, arg in enumerate(invoker_signature.arguments.values()):
                type = DataType.Find(arg.signature)
                if type.RequireTalloc:
                    return True

            return False

    class Interfaces(Base):
        """
            Generator for:

            - interfaces.h
        """

        def __init__(self, header, interfaces):
            super(Generator.Interfaces, self).__init__()

            self.header = header
            self.interfaces = interfaces

        def toggleKeygen(self, tpl, keys, member, sbus_signature):
            """
                Insert a keygen information if this member supports keying.
            """

            if not tpl.hasToggle("keygen"):
                return

            args = InvokerKeygen.GatherKeyArguments(member, sbus_signature)

            if args is None:
                tpl.show("keygen", False)
                return

            tpl.show('keygen', True)

            for idx, arg in args.items():
                type = DataType.Find(arg.signature)

                if type.keyFormat is None:
                    raise ValueError(
                        ('Data type "%s" does not '
                         'support key generator') % type.sbus_type
                    )

                tpl.add('key-argument', {
                    'key-index': idx,
                    'key-format': type.keyFormat
                })

            keys['key-signature'] = sbus_signature.signature

        def toggleAnnotations(self, tpl, annotations):
            """
                Make annotations section visible if there are any to show.
            """

            # We do not include codegen annotations.
            filtered = Generator.FilterAnnotations(annotations)

            show = True
            if filtered is None or not filtered:
                show = False

            tpl.show("annotations", show)

        def setMember(self, template_name, interface, member):
            tpl = self.header.get(template_name)

            invoker = Invoker(member.input, member.output)

            keys = {
                'interface': interface.name,
                'name': member.name,
                'token': self.tokenizeName(interface, member),
                'input-signature': invoker.input.invokerSignature,
                'output-signature': invoker.output.invokerSignature
            }

            if hasattr(member, 'type'):
                keys['type'] = DataType.SBusToDBusType(member.type)

            self.toggleKeygen(tpl, keys, member, member.input)
            self.toggleAnnotations(tpl, member.annotations)
            self.setInputArguments(tpl, member.input)
            self.setOutputArguments(tpl, member.output)

            tpl.set(keys)

        def setInterface(self, template_name, interface):
            tpl = self.header.get(template_name)

            self.toggleAnnotations(tpl, interface.annotations)

            keys = {
                'name': interface.name,
                'token': self.tokenizeName(interface),
            }

            tpl.set(keys)

        def generate(self):
            for interface in self.interfaces.values():
                self.setInterface('interface', interface)

                for method in interface.methods.values():
                    self.setMember('method', interface, method)

                for signal in interface.signals.values():
                    self.setMember('signal', interface, signal)

                for property in interface.properties.values():
                    self.setMember('property', interface, property)

    class Symbols(Base):
        """
            Generator for:

            - symbols.c
            - symbols.h
        """

        def __init__(self, source, header, interfaces):
            super(Generator.Symbols, self).__init__()

            self.source = source
            self.header = header
            self.interfaces = interfaces

        def annotation(self, annotation):
            value = "NULL"
            if annotation.value is not None:
                value = '"%s"' % annotation.value

            return {
                'annotation-name': annotation.name,
                'annotation-value': value
            }

        def argument(self, argument):
            return {
                'arg-type': DataType.SBusToDBusType(argument.signature),
                'arg-name': argument.name
            }

        def generateAnnotations(self, token, annotations):
            # We do not include codegen annotations.
            filtered = Generator.FilterAnnotations(annotations)

            if filtered is None or not filtered:
                return

            tpl = self.source.get('annotations')

            for annotation in filtered.values():
                tpl.add('annotation', self.annotation(annotation))

            keys = {'token': token}
            tpl.set(keys)
            self.header.get('annotations').set(keys)

        def generateMember(self, interface, member, type):
            token = self.tokenizeName(interface, member)
            keys = {'token': token}

            tpl = self.source.get(type)
            for arg in member.input.arguments.values():
                tpl.add('input', self.argument(arg))

            for arg in member.output.arguments.values():
                tpl.add('output', self.argument(arg))

            tpl.set(keys)
            self.header.get(type).set(keys)
            self.generateAnnotations(token, member.annotations)

        def generate(self):
            for interface in self.interfaces.values():
                self.generateAnnotations(self.tokenizeName(interface),
                                         interface.annotations)

                for method in interface.methods.values():
                    self.generateMember(interface, method, 'method')

                for signal in interface.signals.values():
                    self.generateMember(interface, signal, 'signal')

    class Arguments(Base):
        """
            Generator for:

            - arguments.c
            - arguments.h
        """

        def __init__(self, source, header, invoker_arguments):
            super(Generator.Arguments, self).__init__()

            self.source = source
            self.header = header
            self.invoker_arguments = invoker_arguments

        def generate(self):
            self.generateSource()
            self.generateHeader()

        def generateSource(self):
            tpl = self.source.get("arguments")
            for signature, args in self.invoker_arguments.items():
                for idx, arg in enumerate(args.values()):
                    type = DataType.Find(arg.signature)
                    memctx = "mem_ctx, " if type.RequireTalloc else ""
                    keys = {"arg-signature": arg.signature,
                            "talloc-context": memctx,
                            "index": idx}
                    tpl.add('read-argument', keys)
                    tpl.add('write-argument', keys)

                keys = {"signature": signature}
                tpl.set(keys)

        def generateHeader(self):
            tpl = self.header.get("arguments")
            for signature, args in self.invoker_arguments.items():
                for idx, arg in enumerate(args.values()):
                    keys = {"type": DataType.Find(arg.signature).inputCType,
                            "index": idx}
                    tpl.add('args', keys)

                keys = {"signature": signature}
                tpl.set(keys)

    class Invokers(Base):
        """
            Generator for:

            - invokers.c
            - invokers.h
        """

        def __init__(self, source, header, invokers):
            super(Generator.Invokers, self).__init__()

            self.source = source
            self.header = header
            self.invokers = invokers

        def generate(self):
            self.generateSource()
            self.generateHeader()

        def generateSource(self):
            tpl = self.source.get("invoker")
            for invoker in self.invokers.values():
                tpl.show("if-input-arguments",
                         self.hasParsable(invoker.input))
                tpl.show("if-output-arguments",
                         self.hasParsable(invoker.output))

                self.setInputArguments(tpl, invoker.input)
                self.setOutputArguments(tpl, invoker.output)

                keys = {"input-signature": invoker.input.invokerSignature,
                        "output-signature": invoker.output.invokerSignature}

                tpl.set(keys)

        def generateHeader(self):
            tpl = self.header.get("invoker")
            for invoker in self.invokers.values():
                keys = {"input-signature": invoker.input.invokerSignature,
                        "output-signature": invoker.output.invokerSignature}
                tpl.set(keys)

    class Keygens(Base):
        """
            Generator for:

            - keygens.c
            - keygens.h
        """

        def __init__(self, source, header, keygens):
            super(Generator.Keygens, self).__init__()

            self.source = source
            self.header = header
            self.keygens = keygens

        def generate(self):
            for pair in self.keygens.values():
                name = 'key' if pair.arguments else 'key-no-arguments'

                self.set(self.source.get(name), pair.signature, pair.arguments)
                self.set(self.header.get(name), pair.signature, pair.arguments)

        def set(self, tpl, signature, args):
            if args is None:
                return

            for idx, arg in args.items():
                type = DataType.Find(arg.signature)

                if type.keyFormat is None:
                    raise ValueError(
                        ('Data type "%s" does not '
                         'support key generator') % type.sbus_type
                    )

                tpl.add('key-argument', {
                    'key-index': idx,
                    'key-format': type.keyFormat
                })

            tpl.set({'key-signature': signature})

    class Properties(Base):
        """
            Generator for:

            - client_properties.h
        """

        def __init__(self, header, interfaces):
            super(Generator.Properties, self).__init__()

            self.header = header
            self.interfaces = interfaces

        def generate(self):
            tpl = self.header.get("properties")
            for iface in self.interfaces.values():
                if not iface.properties:
                    continue

                added = False

                for property in iface.properties.values():
                    if not property.isReadable():
                        continue

                    if not InvokerCaller.IsWanted(iface, property, "either"):
                        continue

                    added = True

                    type = DataType.Find(property.type)
                    tpl.add("property", {'name': property.name,
                                         'input-type': type.inputCType})

                # Do not generate GetAll caller if we are not interested in
                # properties callers.
                if not added:
                    tpl.clear()
                    continue

                keys = {"token": self.getInterfaceName(iface)}

                tpl.set(keys)

    class MethodCalls(Base):
        """
            Generator for sync and async method callers.
        """

        def __init__(self, source, header, interfaces, type, invokers):
            super(Generator.MethodCalls, self).__init__()

            self.source = source
            self.header = header
            self.interfaces = interfaces
            self.type = type
            self.invokers = invokers

        def generate(self):
            self.generateInvokers()
            self.generateCallers(self.source.get("method-caller"))
            self.generateCallers(self.header.get("method-caller"))

        def generateInvokers(self):
            tpl = self.source.get("method-invoker")
            for invoker in self.invokers.values():
                tpl.show("if-input-arguments",
                         self.hasParsable(invoker.input))
                tpl.show("if-output-arguments",
                         self.hasParsable(invoker.output))
                tpl.show("if-raw-input",
                         self.hasRaw(invoker.input))
                tpl.show("if-raw-output",
                         self.hasRaw(invoker.output))
                tpl.show("if-empty-input",
                         self.hasEmpty(invoker.input))
                tpl.show("if-has-output",
                         self.hasAny(invoker.output))
                tpl.show("if-use-talloc",
                         self.outputNeedTalloc(invoker.output))
                tpl.show("if-show-dummy",
                         self.showDummy(invoker))

                self.setInputArguments(tpl, invoker.input)
                self.setOutputArguments(tpl, invoker.output)

                keys = {"input-signature": invoker.input.invokerSignature,
                        "output-signature": invoker.output.invokerSignature}

                tpl.set(keys)

        def generateCallers(self, tpl):
            for iface in self.interfaces.values():
                for method in iface.methods.values():
                    if not InvokerCaller.IsWanted(iface, method, self.type):
                        continue

                    invoker = Invoker(method.input, method.output)
                    tpl.show("if-raw-input",
                             self.hasRaw(invoker.input))
                    tpl.show("if-raw-output",
                             self.hasRaw(invoker.output))
                    tpl.show("if-has-output",
                             self.hasAny(invoker.output))
                    tpl.show("if-use-talloc",
                             self.outputNeedTalloc(invoker.output))

                    self.setInputArguments(tpl, invoker.input)
                    self.setOutputArguments(tpl, invoker.output)

                    keygen = InvokerKeygen.BuildKeygenName(method,
                                                           method.input)
                    keys = {
                        "token": self.getMemberName(iface, method),
                        "iface": iface.name,
                        "method": method.name,
                        "keygen": keygen,
                        "input-signature": invoker.input.invokerSignature,
                        "output-signature": invoker.output.invokerSignature
                    }

                    tpl.set(keys)

        def showDummy(self, invoker):
            return self.hasEmpty(invoker.output) and not \
                self.hasParsable(invoker.input)

    class SignalCalls(Base):
        """
            Generator for sync and async signal callers.
        """

        def __init__(self, source, header, interfaces, type, invokers):
            super(Generator.SignalCalls, self).__init__()

            self.source = source
            self.header = header
            self.interfaces = interfaces
            self.type = type
            self.invokers = invokers

        def generate(self):
            self.generateInvokers()
            self.generateCallers(self.source.get("signal-caller"))
            self.generateCallers(self.header.get("signal-caller"))

        def generateInvokers(self):
            tpl = self.source.get("signal-invoker")
            for invoker in self.invokers.values():
                tpl.show("if-input-arguments", self.hasParsable(invoker.input))
                tpl.show("if-empty-input", self.hasEmpty(invoker.input))
                tpl.show("if-raw-input", self.hasRaw(invoker.input))

                self.setInputArguments(tpl, invoker.input)

                keys = {"input-signature": invoker.input.invokerSignature}

                tpl.set(keys)

        def generateCallers(self, tpl):
            for iface in self.interfaces.values():
                for signal in iface.signals.values():
                    if not InvokerCaller.IsWanted(iface, signal, self.type):
                        continue

                    invoker = Invoker(signal.input, signal.output)
                    tpl.show("if-raw-input", self.hasRaw(invoker.input))

                    self.setInputArguments(tpl, invoker.input)

                    keys = {"token": self.getMemberName(iface, signal),
                            "iface": iface.name,
                            "signal": signal.name,
                            "input-signature": invoker.input.invokerSignature}

                    tpl.set(keys)

    class PropertyCalls(Base):
        """
            Generator for sync and async property callers.
        """

        def __init__(self, source, header, interfaces, type,
                     get_invokers, set_invokers):
            super(Generator.PropertyCalls, self).__init__()

            self.source = source
            self.header = header
            self.interfaces = interfaces
            self.type = type
            self.get_invokers = get_invokers
            self.set_invokers = set_invokers

        def generate(self):
            self.generateInvokers(self.source.get("get-invoker"),
                                  self.get_invokers)
            self.generateInvokers(self.source.get("set-invoker"),
                                  self.set_invokers)
            self.generateCallers(self.source.get("property-caller"))
            self.generateCallers(self.header.get("property-caller"))
            self.generateGetAll(self.source.get("getall-caller"))
            self.generateGetAll(self.header.get("getall-caller"))

        def generateInvokers(self, tpl, invokers):
            for invoker in invokers.values():
                type = None
                if self.hasAny(invoker.input):
                    type = DataType.Find(
                        invoker.input.arguments['value'].signature
                    )
                elif self.hasAny(invoker.output):
                    type = DataType.Find(
                        invoker.output.arguments['value'].signature
                    )
                else:
                    raise ValueError(
                        'Invoker has no input nor output argument\n'
                    )

                tpl.show("if-use-talloc", type.RequireTalloc)

                keys = {"input-signature": invoker.input.invokerSignature,
                        "output-signature": invoker.output.invokerSignature,
                        "input-type": type.inputCType,
                        "output-type": type.outputCType,
                        "dbus-type": type.dbus_type}

                tpl.set(keys)

        def generateCallers(self, tpl):
            for iface in self.interfaces.values():
                for property in iface.properties.values():
                    if not InvokerCaller.IsWanted(iface, property, self.type):
                        continue

                    if property.isReadable():
                        tpl.show("get", True)

                    if property.isWritable():
                        tpl.show("set", True)

                    invoker = Invoker(property.input, property.output)
                    type = DataType.Find(property.type)

                    tpl.show("get-static", not type.RequireTalloc)
                    tpl.show("get-talloc", type.RequireTalloc)

                    keys = {
                        "token": self.getMemberName(iface, property),
                        "iface": iface.name,
                        "property": property.name,
                        "input-signature": invoker.input.invokerSignature,
                        "output-signature": invoker.output.invokerSignature,
                        "input-type": type.inputCType,
                        "output-type": type.outputCType
                    }

                    tpl.set(keys)

        def generateGetAll(self, tpl):
            for iface in self.interfaces.values():
                if not iface.properties:
                    continue

                added = False

                for property in iface.properties.values():
                    if not property.isReadable():
                        continue

                    if not InvokerCaller.IsWanted(iface, property, self.type):
                        continue

                    added = True

                    type = DataType.Find(property.type)
                    invoker = Invoker(property.input, property.output)

                    loop_name = 'property-static'
                    if type.RequireTalloc:
                        loop_name = 'property-talloc'

                    tpl.add(loop_name, {
                        'name': property.name,
                        'input-type': type.inputCType,
                        'output-type': type.outputCType,
                        'output-signature': invoker.output.invokerSignature
                    })

                # Do not generate GetAll caller if we are not interested in
                # properties callers.
                if not added:
                    tpl.clear()
                    continue

                keys = {"token": self.getInterfaceName(iface),
                        "iface": iface.name}

                tpl.set(keys)



""" Automation of sssctl suite

:requirement: IDM-SSSD-REQ: Status utility
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import pytest
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.exceptions import SSSDException


def client_version(multihost):
    if [int(s) for s in multihost.client[0].distro if s.isdigit()][0] >= 9:
        return True


@pytest.mark.usefixtures('default_sssd')
@pytest.mark.sssctl
class Testsssctl(object):
    """
    This is test case class for sssctl suite
    """
    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__user_show_cache_expiration_time')
    @pytest.mark.tier1_2
    def test_0001_bz1640576(self, multihost,
                            backupsssdconf,
                            localusers):
        """
        :title: IDM-SSSD-TC: sssctl: sssctl reports incorrect
         information about local user's cache entry expiration time
        :id: 9315c119-8c69-4685-836d-0f71b5d0684c
        """
        if not multihost.client[0].detect_files_provider():
            pytest.skip("Files Provider support isn't available, skipping")
        users = localusers
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        tools.remove_sss_cache('/var/log/sssd')
        sssd_param = {'domains': 'local'}
        tools.sssd_conf('sssd', sssd_param)
        param = {'id_provider': 'files',
                 'passwd_files': '/etc/passwd'}
        tools.sssd_conf('domain/local', param)
        multihost.client[0].service_sssd('start')
        for user in users.keys():
            sssctl_cmd = 'sssctl user-show %s' % user
            cmd = multihost.client[0].run_command(sssctl_cmd,
                                                  raiseonerr=False)
            assert 'Cache entry expiration time: Never'\
                   in cmd.stdout_text

    @pytest.mark.converted('test_sssctl.py', 'test_sssctl__handle_implicit_domain')
    @pytest.mark.tier1_2
    def test_0002_bz1599207(self, multihost,
                            backupsssdconf,
                            localusers):
        """
        :title: IDM-SSSD-TC: sssctl: sssd tools do not handle the implicit
         domain
        :id: b5ff4e8f-ce9f-4731-bbaa-bf2a8425dc15
        """
        if not multihost.client[0].detect_files_provider():
            pytest.skip("Files Provider support isn't available, skipping")
        users = localusers
        tools = sssdTools(multihost.client[0])
        multihost.client[0].service_sssd('stop')
        tools.remove_sss_cache('/var/lib/sss/db')
        tools.remove_sss_cache('/var/log/sssd')
        tools.sssd_conf("sssd",
                        {'enable_files_domain': 'true'},
                        action='update')
        multihost.client[0].service_sssd('start')
        for user in users.keys():
            cmd = multihost.client[0].run_command('getent'
                                                  ' -s sss'
                                                  ' passwd %s '
                                                  '&& sssctl '
                                                  'user-show %s' %
                                                  (user, user),
                                                  raiseonerr=False)
            assert 'Cache entry creation date' in \
                   cmd.stdout_text and cmd.returncode == 0

    @pytest.mark.converted('test_sss_cache.py', 'test_sss_cache__cache_expire_message')
    @pytest.mark.tier1_2
    def test_0003_bz1661182(self, multihost,
                            backupsssdconf):
        """
        :title: sss_cache prints spurious error messages
         when invoked from shadow-utils on package install
        :id: 8f2868d2-1ece-11ec-ac6d-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1661182
        :steps:
          1. Configure sssd without any domain
          2. Restart sssd (sssd should not be running after this)
          3. Modify existing local user usermod -a -G wheel user1
          4. This message
             '[sss_cache] [confdb_get_domains] (0x0010):
             No domains configured, fatal error!'
             must not appear in console
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
        """
        if not multihost.client[0].detect_files_provider():
            pytest.skip("Files Provider support isn't available, skipping")
        tools = sssdTools(multihost.client[0])
        ldap_params = {'enable_files_domain': 'false'}
        tools.sssd_conf('sssd', ldap_params)
        with pytest.raises(SSSDException):
            multihost.client[0].service_sssd('restart')
        ps_cmd = "> /var/log/sssd/sssd.log"
        multihost.client[0].run_command(ps_cmd)
        ps_cmd = "useradd user1_test"
        multihost.client[0].run_command(ps_cmd, raiseonerr=False)
        ps_cmd = "usermod -a -G wheel user1_test"
        cmd = multihost.client[0].run_command(ps_cmd)
        assert 'No domains configured, fatal error!' \
               not in cmd.stdout_text
        ps_cmd = "userdel user1_test"
        multihost.client[0].run_command(ps_cmd)
        for ps_cmd in ('sss_cache -U',
                       'sss_cache -G',
                       'sss_cache -E',
                       'sss_cache -u non-existinguser'):
            cmd = multihost.client[0].run_command(ps_cmd)
            assert 'No domains configured, fatal error!' \
                   not in cmd.stdout_text



#
#   Authors:
#       Pavel Brezina <pbrezina@redhat.com>
#
#   Copyright (C) 2017 Red Hat
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from collections import OrderedDict
import xml.etree.ElementTree as etree


class Introspectable:
    class Element(object):
        """ This is a basic introspectable object. This class will make
            sure that the given xml element is of correct type and provide
            some helper functions to simplify work of the children.

            Children objects must implement TagName attribute, which contains
            the name of the expected xml tag.

            All introspectable objects contain the following properties:
            - name : str -- name of the object
            - annotations : OrderedDict -- available annotations
        """
        def __init__(self, element):
            self.check(element, self.TagName)

            self.element = element
            self.name = element.attrib["name"]
            self.annotations = self.find(SBus.Annotation)

        def find(self, object_class):
            return Introspectable.FindElements(self.element, object_class)

        def check(self, element, tagname):
            if element.tag != tagname:
                raise ValueError('Unexpected tag name "%s" (%s expected)!'
                                 % (element.tag, tagname))
            if "name" not in element.attrib:
                raise ValueError('Missing attribute name!')

        def getAttr(self, name, default_value):
            return self.element.attrib.get(name, default_value)

        def getExistingAttr(self, name):
            if name not in self.element.attrib:
                raise ValueError('Element %s name="%s" is missing attribute %s'
                                 % (self.TagName, self.name, name))

            return self.element.attrib[name]

    class Invokable(Element):
        """ This is a base class for invokable objects -- methods and signals.
            Invokable objects has available additional attributes:

            - input OrderedDict -- input signature and arguments
            - output : OrderedDict -- output signature and arguments
        """
        def __init__(self, element):
            super(Introspectable.Invokable, self).__init__(element)

            self.key = self.getAttr("key", None)

            self.arguments = self.find(SBus.Argument)
            input = self.getInputArguments()
            output = self.getOutputArguments()

            self.input = SBus.Signature(input, self.annotations)
            self.output = SBus.Signature(output, self.annotations)
            return

        def getInputArguments(self):
            return self.getArguments("in")

        def getOutputArguments(self):
            return self.getArguments("out")

        def getArguments(self, type):
            args = OrderedDict()

            for name, arg in self.arguments.items():
                if type == "in" and arg.isInput():
                    args[name] = arg
                    continue
                if type == "out" and arg.isOutput():
                    args[name] = arg
                    continue

            return args

    @staticmethod
    def Introspect(path):
        root = etree.parse(path).getroot()
        return Introspectable.FindElements(root, SBus.Interface)

    @staticmethod
    def FindElements(parent, object_class):
        dict = OrderedDict()
        for child in parent:
            if child.tag != object_class.TagName:
                continue
            object = object_class(child)

            if object.name in dict:
                raise ValueError('%s name="%s" is already present '
                                 'in the same parent element\n'
                                 % (object_class.TagName, object.name))

            dict[object.name] = object

        """
        Arguments can't be sorted and annotations order should be left on
        the author of introspection. Otherwise we want to sort the dictionary
        alphabetically based on keys.
        """
        if object_class in [SBus.Argument, SBus.Annotation]:
            return dict

        return OrderedDict(sorted(dict.items()))


class SBus:
    class Interface(Introspectable.Element):
        TagName = "interface"

        def __init__(self, element):
            super(SBus.Interface, self).__init__(element)

            self.methods = self.find(SBus.Method)
            self.signals = self.find(SBus.Signal)
            self.properties = self.find(SBus.Property)
            return

    class Method(Introspectable.Invokable):
        TagName = "method"

        def __init__(self, element):
            super(SBus.Method, self).__init__(element)

    class Signal(Introspectable.Invokable):
        TagName = "signal"

        def __init__(self, element):
            super(SBus.Signal, self).__init__(element)

    class Property(Introspectable.Invokable):
        TagName = "property"

        def __init__(self, element):
            self.name = element.attrib["name"]
            self.element = element
            self.access = self.getExistingAttr("access")
            self.type = self.getExistingAttr("type")

            super(SBus.Property, self).__init__(element)

            if self.key is not None:
                raise ValueError('Keying is not supported on properties: %s '
                                 % self.name)

        def getInputArguments(self):
            if not self.isWritable():
                return {}

            return {"value": SBus.Argument.Create("value", self.type, "in")}

        def getOutputArguments(self):
            if not self.isReadable():
                return {}

            return {"value": SBus.Argument.Create("value", self.type, "out")}

        def isReadable(self):
            return self.access == "read" or self.access == "readwrite"

        def isWritable(self):
            return self.access == "write" or self.access == "readwrite"

    class Annotation(Introspectable.Element):
        TagName = "annotation"

        def __init__(self, element):
            super(SBus.Annotation, self).__init__(element)

            self.value = self.getAttr("value", None)
            return

        @staticmethod
        def Find(annotations, name, default_value):
            if name in annotations:
                annotation = annotations[name]
                if annotation.value is None:
                    return default_value
                return annotation.value
            return default_value

        @staticmethod
        def FindBool(annotations, name, Assume=False):
            assume = "true" if Assume else "false"
            value = SBus.Annotation.Find(annotations, name, assume)
            if value.lower() == "true":
                return True
            else:
                return False

        @staticmethod
        def CheckIfTrue(names, annotations):
            for name in names:
                if SBus.Annotation.FindBool(annotations, name, False):
                    return True

            return False

        @staticmethod
        def CheckIfFalse(names, annotations):
            for name in names:
                if not SBus.Annotation.FindBool(annotations, name, True):
                    return False

            return True

        @staticmethod
        def AtleastOneIsSet(names, annotations):
            for name in names:
                value = SBus.Annotation.Find(annotations, name, None)
                if value is not None:
                    return True

            return False

    class Argument(Introspectable.Element):
        TagName = "arg"

        def __init__(self, element, Name=None, Type=None, Direction=None,
                     Key=None):
            if element is None:
                self.element = None
                self.name = Name
                self.signature = Type
                self.direction = Direction
                self.key = Key
                return

            super(SBus.Argument, self).__init__(element)

            self.signature = self.getExistingAttr("type")
            self.direction = self.getAttr("direction", "in")
            self.key = self.getAttr("key", None)

        def isInput(self):
            return self.direction == "in"

        def isOutput(self):
            return not self.isInput()

        @staticmethod
        def Create(name, type, direction):
            return SBus.Argument(element=None,
                                 Name=name,
                                 Type=type,
                                 Direction=direction)

    class Signature:
        def __init__(self, args, annotations):
            self.annotations = annotations
            self.signature = self.getSignature(args)
            self.arguments = args

        def getSignature(self, args):
            signature = ""
            for arg in args.values():
                signature += arg.signature

            return signature



"""Automation for krb ldap connection

:requirement: krb_ldap_connection
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import time
import pytest
from sssd.testlib.common.utils import sssdTools
from constants import ds_instance_name


@pytest.mark.krbldapconnection
@pytest.mark.tier1_3
@pytest.mark.usefixtures('setup_sssd_gssapi',
                         'create_posix_usersgroups')
class TestKrbConnectionTimeout(object):

    def test_timeoutkrb(self, multihost):
        """
        :title: IDM-SSSD-TC: krb_provider: krb_ldap_connection:
         Test for krb
        :id: b8b3e2ab-d24e-4da8-b750-e405f25c6908
        :expectedresults: In this case value of timeout depends upon
         max value of lifetime of ticket that is in our case is 120 seconds
         (2 mins). After that it should have to release the connection
        """
        cmd_mod_user = ["kadmin.local", "-q", "modprinc -maxlife 2mins foo1"]
        multihost.master[0].run_command(cmd_mod_user)
        cmd_mod = ["kadmin.local", "-q", "modprinc -maxlife 2mins krbtgt/EXAMPLE.TEST"]
        multihost.master[0].run_command(cmd_mod)
        multihost.client[0].log.info(
            '\n\n\nTesting for the case where timeout value is'
            'out of range (of integer). ldap_connection_expire_'
            'timeout = krb')
        domain_params = {'ldap_connection_expire_timeout': None}
        sssdTools(
            multihost.client[0]).sssd_conf(
            'domain/%s' %
            (ds_instance_name), domain_params, 'delete')
        cmd_restart = "systemctl restart sssd"
        multihost.client[0].run_command(cmd_restart)
        logfile = '/var/log/sssd/sssd_%s.log' % ds_instance_name
        sssdTools(multihost.client[0]).clear_sssd_cache()
        timeout = 120
        lookup_u = 'getent passwd foo1@%s' % ds_instance_name
        cmd = multihost.client[0].run_command(lookup_u)
        assert cmd.returncode == 0
        time.sleep(timeout + 5)
        lookup_u = 'getent passwd foo2@%s' % ds_instance_name
        cmd = multihost.client[0].run_command(lookup_u)
        assert cmd.returncode == 0
        string = "Connection is about to expire, releasing it"
        file_content = multihost.client[0].get_file_contents(logfile)
        x = string.encode('utf-8') in file_content
        if x is True:
            assert True
        else:
            assert False
        cmd_remove_log = "rm /var/log/sssd/sssd_example1.log"
        multihost.client[0].run_command(cmd_remove_log)
        cmd_mod_user = ["kadmin.local", "-q", "modprinc -maxlife 1day foo1"]
        multihost.master[0].run_command(cmd_mod_user)
        cmd_mod = ["kadmin.local", "-q", "modprinc -maxlife 1day krbtgt/EXAMPLE.TEST"]
        multihost.master[0].run_command(cmd_mod)



#
#   Authors:
#       Pavel Brezina <pbrezina@redhat.com>
#
#   Copyright (C) 2017 Red Hat
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import re
import errno
import textwrap
import os.path


class Template:
    def __init__(self, name, templateFile, template):
        template = self.removeLines(template)

        self.templateFile = templateFile
        self.name = name
        self.loops = {}
        self.toggles = {}
        self.template = self.parse(template)
        self.output = ""

    def parse(self, template):
        template = self.parseLoops(template)
        template = self.parseToggles(template)
        return template

    def parseLoops(self, template):
        template = self.Pattern.Loop.sub(self.processLoops, template)
        return self.Pattern.LoopLine.sub(self.processLoops, template)

    def processLoops(self, match):
        name = match.group(1)
        template = self.removeLines(match.group(2))
        index = 0

        if name not in self.loops:
            self.loops[name] = self.Loop()

        index = self.loops[name].addTemplate(template)
        return '$${loop:%s:%d}' % (name, index)

    def parseToggles(self, template):
        template = self.Pattern.Toggle.sub(self.processToggles, template)
        return self.Pattern.ToggleLine.sub(self.processToggles, template)

    def processToggles(self, match):
        name = match.group(1)
        if_visible = self.removeLines(match.group(2))
        if_hidden = self.removeLines(match.group(4))
        index = 0

        if name not in self.toggles:
            self.toggles[name] = self.Toggle()

        index = self.toggles[name].addTemplate(if_visible, if_hidden)
        return '$${toggle:%s:%d}' % (name, index)

    def add(self, loop_name, values):
        """Add new item into <loop name="$loop_name"> template.
           Setting its attributes to $values.
        """
        if loop_name not in self.loops:
            return self
        self.loops[loop_name].set(values)
        return self

    def show(self, toggle_name, isVisible):
        """Make <toggle name="$toggle_name"> either visible or hidden
           within the template.
        """
        if not self.hasToggle(toggle_name):
            return

        self.toggles[toggle_name].show(isVisible)

    def hasToggle(self, name):
        return name in self.toggles

    def hasLoop(self, name):
        return name in self.loops

    def set(self, values):
        """Set template attributes to $values, push generated content into
           the output file and reset this template.
        """
        template = self.template
        for key, toggle in self.toggles.items():
            for idx, toggletpl in enumerate(toggle.templates):
                pattern = "$${toggle:%s:%d}" % (key, idx)
                template = template.replace(pattern, toggletpl.generate())

        self.output = self.Set(template, values)
        self.templateFile.push(self.generate())
        self.clear()

    def pushOriginal(self):
        """Push original template into the output file
        """
        self.templateFile.push(self.template)

    def clear(self):
        for loop in self.loops.values():
            loop.clear()

        for toggle in self.toggles.values():
            toggle.show(False)

        self.output = ""

    def generate(self):
        output = self.output
        for key, loop in self.loops.items():
            for idx, content in enumerate(loop.templates):
                pattern = "$${loop:%s:%d}" % (key, idx)
                output = output.replace(pattern, loop.get(idx), 1)
        return output

    @staticmethod
    def Set(content, values):
        output = content
        for key, value in values.items():
            output = output.replace("${" + key + "}", str(value))
        return output

    def removeLines(self, content):
        """Remove unneeded lines and spaces. There are some additional lines
           and spaces that may end up in the template after parsing. This
           method will remove new line after <@template-tag> and spaces
           from otherwise empty lines.
        """
        if content is None:
            return content

        content = self.Pattern.NewLine.sub('', content, 1)
        content = self.Pattern.EmptyLine.sub('', content)
        return content

    class Pattern:
        Template = re.compile(
            r' *<template name="(\S+)">(.*?)</template>\r?\n?',
            re.MULTILINE | re.DOTALL
        )

        Loop = re.compile(
            r' *<loop name="(\S+)">(.*?)</loop>\r?\n?',
            re.MULTILINE | re.DOTALL
        )

        LoopLine = re.compile(
            r'<loop line name="(\S+)">(.*?)</loop>',
            re.MULTILINE | re.DOTALL
        )

        Toggle = re.compile(
            r' *<toggle name="(\S+)">(.*?)(<or>(.*?))?</toggle>\r?\n?',
            re.MULTILINE | re.DOTALL
        )

        ToggleLine = re.compile(
            r'<toggle line name="(\S+)">(.*?)(<or>(.*?))?</toggle>',
            re.MULTILINE | re.DOTALL
        )

        NewLine = re.compile('^\r?\n')

        EmptyLine = re.compile('^ *$', re.MULTILINE)

    class Loop:
        def __init__(self):
            self.templates = []
            self.num_templates = 0

        def addTemplate(self, template):
            self.templates.append(self.LoopTemplate(template))
            self.num_templates += 1
            return self.num_templates - 1

        def set(self, values):
            for template in self.templates:
                template.set(values)

        def clear(self):
            for template in self.templates:
                template.clear()

        def get(self, index):
            return self.templates[index].generate()

        class LoopTemplate:
            def __init__(self, template):
                self.template = template
                self.output = ""

            def set(self, values):
                self.output += Template.Set(self.template, values)

            def clear(self):
                self.output = ""

            def generate(self):
                return self.output

    class Toggle:
        def __init__(self):
            self.templates = []
            self.num_templates = 0
            self.visible = False

        def addTemplate(self, if_visible, if_hidden):
            toggletpl = self.ToggleTemplate(self, if_visible, if_hidden)
            self.templates.append(toggletpl)
            self.num_templates += 1
            return self.num_templates - 1

        def show(self, isVisible):
            self.visible = isVisible

        class ToggleTemplate:
            def __init__(self, toggle, if_visible, if_hidden):
                self.toggle = toggle
                self.if_visible = if_visible
                self.if_hidden = if_hidden

            def generate(self):
                if self.toggle.visible:
                    return self.if_visible
                elif self.if_hidden is not None:
                    return self.if_hidden

                return ''


class TemplateFile:
    """Parse file contents into templates.

       Obtain template with .get and set its content. When all the content is
       set, you can call .generate to obtain generated content or .write
       to write it to a file.
    """

    def __init__(self, path):
        with open(path, "r") as file:
            contents = file.read()

        self.templates = {}
        self.output = ""
        self.parse(contents)

    def parse(self, template):
        for (name, content) in Template.Pattern.Template.findall(template):
            content = textwrap.dedent(content)
            self.templates[name] = Template(name, self, content)

    def get(self, name):
        return self.templates[name]

    def has(self, name):
        return name in self.templates

    def push(self, content):
        self.output += content

    def generate(self):
        return self.output

    def write(self, filename, postprocess=None):
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except OSError as exception:
                if exception.errno == errno.EEXIST and os.path.isdir(filename):
                    pass
                else:
                    raise

        output = self.generate().rstrip() + '\n'
        if postprocess is not None:
            output = postprocess(output)

        if not self.needsOverride(filename, output):
            return

        with open(filename, "w") as file:
            file.write(output)

    def needsOverride(self, filename, content):
        """
            Do not override the file unless it is not yet present or its
            current content differs from the generated one. This ensure
            that the file is in correct state and yet it is not rebuild
            during make unless necessary.
        """
        if not os.path.isfile(filename):
            return True

        with open(filename, "r") as file:
            current_content = file.read()
            if current_content != content:
                return True

        return False

    def __str__(self):
        return self.generate()



""" Automation of sanity/journald suite

:requirement: IDM-SSSD-REQ : Journal based logging
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import pytest
import re
import time
from constants import ds_instance_name


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups',
                         'write_journalsssd')
@pytest.mark.journald
class TestJournald(object):
    """
    This is test case class for sanity/journald suite
    """
    @pytest.mark.tier1
    def test_0001_bz1115508(self, multihost, enable_multiple_responders):
        """
        :title: IDM-SSSD-TC: sanity: journald: Send debug logs to journald by
         default bz1115508
        :id: 649107d8-7457-4da3-91c7-32b5bb493fb7
        """
        date = "date" + " --rfc-3339=ns"
        cmd = multihost.client[0].run_command(date, raiseonerr=False)
        date_org = cmd.stdout_text
        date = '"' + date_org[0:19] + '"'

        up_res = {}
        responder = {'jctl_sssd_be': 'sssd_be',
                     'jctl_sssd_pam': 'sssd_pam',
                     'jctl_sssd_nss': 'sssd_nss',
                     'jctl_sssd_autofs': 'sssd_autofs',
                     'jctl_sssd_ssh': 'sssd_ssh',
                     'jctl_sssd_pac': 'sssd_pac',
                     'jctl_sssd_ifp': 'sssd_ifp',
                     'jctl_sssd_sudo': 'sssd_sudo'}

        cmd_lookup = 'getent passwd foo1@%s' % ds_instance_name
        test = multihost.client[0].run_command(cmd_lookup, raiseonerr=False)
        if test.stdout_text.find("foo1") != -1:
            status = 'PASS'
        else:
            status = 'FAIL'
        time.sleep(10)
        multihost.client[0].service_sssd('restart')
        for x in responder:
            cmd = "journalctl --since %s -u sssd SYSLOG_IDENTIFIER=%s" % (
                date, responder[x])
            up_res[x] = cmd
            if 'nss' in up_res[x]:
                up_res[x] = multihost.client[0].run_command(up_res[x],
                                                            raiseonerr=False)
                log = re.compile(r'foo1.*')
                if log.search(up_res[x].stdout_text):
                    status = 'PASS'
                else:
                    status = "FAIL"
            else:
                up_res[x] = multihost.client[0].run_command(up_res[x],
                                                            raiseonerr=False)
                log1 = re.compile(r'.No\sentries.*')
                if log1.search(up_res[x].stdout_text):
                    status = 'FAIL'
                else:
                    status = 'PASS'
        assert status == 'PASS'

    @pytest.mark.tier1
    def test_0002_bz1460724(self, multihost, enable_multiple_responders):
        """
        :title: IDM-SSSD-TC: sanity: journald: SYSLOG_IDENTIFIER is different
        :id: c30e6899-6c6b-42bb-a9a3-4aacd589908e
        """
        pgrep = multihost.client[0].run_command(["pgrep", "-af", "sssd"],
                                                raiseonerr=False)

        if '--logger=journald' in pgrep.stdout_text:
            for resp in ['sssd_be', 'sssd_pam', 'sssd_nss', 'sssd_sudo',
                         'sssd_autofs', 'sssd_ssh', 'sssd_ifp', 'sssd_pac']:
                cmd = "journalctl -r --output=json SYSLOG_IDENTIFIER=%s >" \
                      " /root/op.json" % resp
                cmd = multihost.client[0].run_command(cmd)
                assert cmd.returncode == 0
                file_name = '/root/op.json'
                output = multihost.client[0].get_file_contents(file_name)
                search_str = re.compile(r'.SYSLOG_IDENTIFIER.*.:.*.%s.' % resp)
                assert search_str.search(output.decode())
                rm = 'rm -f /root/op.json'
                multihost.client[0].run_command(rm)



""" Automation of proxy provider suite

:requirement: IDM-SSSD-REQ : Proxy Provider
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
"""
from __future__ import print_function
import pytest
import subprocess
import time
from sssd.testlib.common.utils import sssdTools, LdapOperations
from sssd.testlib.common.ssh2_python import check_login_client
from constants import ds_suffix, ds_instance_name


def execute_cmd(multihost, command):
    """ Execute command on client """
    cmd = multihost.client[0].run_command(command)
    return cmd


@pytest.fixture(scope='class')
def ldap_objects_sssd_client(multihost, request):
    """
        Configure sssd.conf
        Create a dedicated user with
        a DN starting e.g. with cn=...
        Create a local user foo12
    """
    ldap_uri = 'ldap://%s' % multihost.master[0].sys_hostname
    ds_rootdn = 'cn=Directory Manager'
    ds_rootpw = 'Secret123'
    ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
    user_info = {'objectClass': [b'top', b'organizationalUnit'],
                 'ou': 'Users'.encode('utf-8')}
    users_dn = 'ou=Users,dc=example,dc=test'
    (_, _) = ldap_inst.add_entry(user_info, users_dn)
    user_info = {'cn': 'User_CS2'.encode('utf-8'),
                 'homeDirectory': '/home/User_CS2'.encode('utf-8'),
                 'objectClass': [b'account',
                                 b'posixAccount',
                                 b'extensibleObject'],
                 'uidNumber': '1111112'.encode('utf-8'),
                 'gidNumber': '1111112'.encode('utf-8'),
                 'loginShell': '/bin/bash'.encode('utf-8'),
                 'userPassword': 'Secret123'.encode('utf-8'),
                 'uid': ['User_CS2'.encode('utf-8'),
                         'User_CS2_Alias'.encode('utf-8')]}
    user_dn = f'uid=User_CS2,ou=Users,{ds_suffix}'
    (_, _) = ldap_inst.add_entry(user_info, user_dn)
    group_info = {'cn': ['User_CS2_grp1'.encode('utf-8'),
                         'User_CS2_grp1_Alias'.encode('utf-8')],
                  'objectClass': [b'extensibleObject', b'groupOfNames'],
                  'gidNumber': '1111112'.encode('utf-8'),
                  'member': f'uid=User_CS2,ou=Users,{ds_suffix}'.encode('utf-8')}
    group_dn = f'cn=User_CS2_grp1,ou=Groups,{ds_suffix}'
    (_, _) = ldap_inst.add_entry(group_info, group_dn)
    execute_cmd(multihost, "> /etc/pam_ldap.conf")
    execute_cmd(multihost, f"echo 'base {ds_suffix}' > "
                           f"/etc/pam_ldap.conf")
    execute_cmd(multihost, "echo 'pam_password md5' >>"
                           " /etc/pam_ldap.conf")
    execute_cmd(multihost, f"echo 'host {multihost.master[0].ip}' "
                           f">> /etc/pam_ldap.conf")
    execute_cmd(multihost, "echo 'tls_cacertfile "
                           "/etc/openldap/certs/cacert.asc' >>"
                           " /etc/pam_ldap.conf")
    execute_cmd(multihost, "echo 'filter group  "
                           "(objectClass=groupOfNames)'>> "
                           "/etc/nslcd.conf")
    """
        Interaction between nss-pam-ldapd and a feature of pam_usertype.
        pam_usertype might send a lookup for a user called
        'pam_usertype_non_existent:' and nss-pam-ldapd does not like the ':' in
        the username.

        To fix this we need to add following for RHEL9
    """
    if '9' or "Fedora" in execute_cmd(multihost, "cat /etc/redhat-release").stdout_text:
        execute_cmd(multihost, "echo 'validnames /^[a-z0-9._@$()]([a-z0-9._@$()"
                               " \\~-]*[a-z:0-9._@$()~-])?$/i' >> /etc/nslcd.conf")
    execute_cmd(multihost, 'systemctl restart nslcd')

    def del_user_grp():
        """ Restore sssd.conf """
        ldap_inst.del_dn(user_dn)
        ldap_inst.del_dn(group_dn)
        ldap_inst.del_dn(users_dn)
        execute_cmd(multihost, "rm -vf /etc/pam_ldap.conf")

    request.addfinalizer(del_user_grp)


@pytest.mark.usefixtures('setupds',
                         'default_sssd',
                         'sssdproxyldap',
                         'install_nslcd',
                         'ldap_objects_sssd_client')
@pytest.mark.tier1_4
class TestProxyrfc2307bis(object):
    """
    This is test case class for proxy provider suite
    """
    def test_lookup_user_group_netgroup(self, multihost, backupsssdconf):
        """
        :title: case sensitive is true lookup user group
        :id: d0531f56-43e8-11ed-9742-845cf3eff344
        :steps:
          1. While case_sensitive = true, lookup user group
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'true',
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["id User_CS2 | grep User_CS2_grp1",
                  "id User_CS2_Alias | grep User_CS2_grp1",
                  "getent passwd User_CS2 | grep User_CS2",
                  "getent passwd User_CS2_Alias | grep User_CS2",
                  "getent group User_CS2_grp1 | grep User_CS2_grp1",
                  "getent group User_CS2_grp1 | grep User_CS2",
                  "getent group User_CS2_grp1_Alias | grep User_CS2_grp1",
                  "getent group User_CS2_grp1_Alias | grep User_CS2"]:
            execute_cmd(multihost, i)
        for user in ['User_CS2', 'User_CS2_Alias']:
            check_login_client(multihost, user, 'Secret123')
        for i in ["getent passwd user_cs2",
                  "getent passwd user_cs2_alias",
                  "getent group user_cs2_grp1",
                  "getent group user_cs2_grp1"]:
            with pytest.raises(subprocess.CalledProcessError):
                execute_cmd(multihost, i)

    def test_allow_groups_User_CS2_grp1(self, multihost, backupsssdconf):
        """
        :title: simple allow groups is User CS2 grp1
        :id: d798c644-43e8-11ed-bfd5-845cf3eff344
        :steps:
          1. While case_sensitive = true, allow groups User CS2 grp1
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_allow_groups': 'User_CS2_grp1,User_CS2_grp1_Alias',
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent group User_CS2_grp1", "id User_CS2"]:
            execute_cmd(multihost, i)
        check_login_client(multihost, "User_CS2", 'Secret123')

    def test_allow_groups_user_cs2_grp1(self, multihost, backupsssdconf):
        """
        :title: simple allow groups is user cs2 grp1
        :id: de49b3cc-43e8-11ed-a35c-845cf3eff344
        :caseposneg: negative
        :steps:
          1. While case_sensitive = true, allow groups  user cs2 grp1
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_allow_groups': 'user_cs2_grp1',
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent group User_CS2_grp1",
                  "id User_CS2",
                  "> /var/log/secure"]:
            execute_cmd(multihost, i)
        with pytest.raises(Exception):
            check_login_client(multihost, "User_CS2", 'Secret123')
        time.sleep(3)
        execute_cmd(multihost, 'cat /var/log/secure | grep "Access denied for user User_CS2"')

    def test_allow_groups_User_CS2(self, multihost, backupsssdconf):
        """
        :title: simple allow users is User CS2
        :id: e3cd00ce-43e8-11ed-97b6-845cf3eff344
        :steps:
          1. While case_sensitive = true, allow users User CS2
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_allow_users': "User_CS2,User_CS2_Alias",
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent passwd User_CS2", "id User_CS2"]:
            execute_cmd(multihost, i)
        check_login_client(multihost, "User_CS2", 'Secret123')

    def test_allow_groups_user_cs2(self, multihost, backupsssdconf):
        """
        :title: simple allow users is user cs2
        :id: e9ffc58a-43e8-11ed-8df5-845cf3eff344
        :caseposneg: negative
        :steps:
          1. While case_sensitive = true, allow users user cs2
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'true',
                         'access_provider': 'simple',
                         'simple_allow_groups': 'user_cs2',
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for i in ["getent group User_CS2_grp1",
                  "id User_CS2",
                  "> /var/log/secure"]:
            execute_cmd(multihost, i)
        with pytest.raises(Exception):
            check_login_client(multihost, "User_CS2", 'Secret123')
        time.sleep(3)
        execute_cmd(multihost, 'cat /var/log/secure | grep "Access denied for user User_CS2"')

    def test_case_sensitive(self, multihost, backupsssdconf):
        """
        :title: case sensitive is false lookup user group
        :id: f0389800-43e8-11ed-af2a-845cf3eff344
        :steps:
          1. While case_sensitive = false, lookup user group
        :expectedresults:
          1. Should succeed
        """
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'false',
                         'simple_allow_users': 'user_cs2',
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        execute_cmd(multihost, 'echo "ignorecase yes" >> /etc/nslcd.conf')
        execute_cmd(multihost, 'systemctl restart nslcd')
        tools.clear_sssd_cache()
        for i in ["getent passwd user_cs2 | grep user_cs2",
                  "getent passwd User_CS2 | grep user_cs2",
                  "getent passwd user_cs2_alias | grep user_cs2",
                  "getent passwd User_CS2_Alias | grep user_cs2",
                  "getent group User_CS2_grp1 | grep user_cs2_grp1",
                  "getent group user_cs2_grp1 | grep user_cs2",
                  "getent group User_CS2_grp1_Alias | grep user_cs2_grp1",
                  "getent group user_cs2_grp1_alias | grep user_cs2",
                  "id User_cs2 | grep user_cs2_grp1",
                  "id user_cs2_Alias | grep user_cs2_grp1"]:
            execute_cmd(multihost, i)
        client_hostip = multihost.client[0].ip
        for user in ['user_cs2', 'user_cs2_alias']:
            check_login_client(multihost, user, 'Secret123')

    def test_simple_deny_users_user_cs2(self, multihost, backupsssdconf):
        """
        :title: case sensitive is false simple deny users is user cs2
        :id: f60de2ee-43e8-11ed-96e6-845cf3eff344
        :steps:
          1. While case_sensitive = false, deny users user cs2
        :expectedresults:
          1. Should succeed
        """
        client_hostip = multihost.client[0].ip
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'false',
                         'access_provider': 'simple',
                         'simple_deny_users': 'user_cs2',
                         'use_fully_qualified_names': 'False'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for user in ['User_cs2', 'user_cs2_alias']:
            execute_cmd(multihost, "> /var/log/secure")
            with pytest.raises(Exception):
                check_login_client(multihost, user, 'Secret123')
            time.sleep(3)
            execute_cmd(multihost, f'cat /var/log/secure | grep "Access denied for user {user}"')
        execute_cmd(multihost, "> /var/log/secure")
        execute_cmd(multihost, "sed -i 's/user_cs2/user_cs2_alias/' /etc/sssd/sssd.conf")
        tools.clear_sssd_cache()
        for user in ['user_cs2_alias', 'User_cs2']:
            check_login_client(multihost, user, 'Secret123')

    def test_simple_deny_groups_user_cs2_grp1(self, multihost, backupsssdconf):
        """
        :title: case sensitive is false simple deny groups is user cs2 grp1
        :id: fbd57926-43e8-11ed-a640-845cf3eff344
        :steps:
          1. While case_sensitive = false, deny groups user cs2 grp1
        :expectedresults:
          1. Should succeed
        """
        client_hostip = multihost.client[0].ip
        tools = sssdTools(multihost.client[0])
        sssd_params = {'domains': ds_instance_name}
        tools.sssd_conf('sssd', sssd_params)
        domain_name = tools.get_domain_section_name()
        domain_params = {'debug_level': '9',
                         'id_provider': 'proxy',
                         'proxy_lib_name': 'ldap',
                         'proxy_pam_target': 'sssdproxyldap',
                         'case_sensitive': 'false',
                         'access_provider': 'simple',
                         'simple_deny_groups': 'user_cs2_grp1',
                         'use_fully_qualified_names': 'false'}
        tools.sssd_conf('domain/' + domain_name, domain_params)
        tools.clear_sssd_cache()
        for user in ['User_cs2', 'user_cs2_alias']:
            execute_cmd(multihost, "> /var/log/secure")
            with pytest.raises(Exception):
                check_login_client(multihost, user, 'Secret123')
            time.sleep(3)
            execute_cmd(multihost, f'cat /var/log/secure | grep "Access denied for user {user}"')
        execute_cmd(multihost, "> /var/log/secure")
        execute_cmd(multihost, "sed -i 's/user_cs2_grp1/user_cs2_grp1_alias/' /etc/sssd/sssd.conf")
        tools.clear_sssd_cache()
        for user in ['user_cs2_alias', 'User_cs2']:
            check_login_client(multihost, user, 'Secret123')



""" AD-Provider BZ Automations

:requirement: ad_parameters
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import time
import random
import re
import pytest

from sssd.testlib.common.utils import sssdTools


@pytest.mark.adparameters
class TestBugzillaAutomation(object):
    """ BZ Automated Test Cases """

    @pytest.mark.tier1
    def test_0001_bz1296618(self, multihost, adjoin,
                            create_aduser_group):
        """
        :title: IDM-SSSD-TC: ad_provider: ad_parameters: Properly remove
         OriginalMemberOf attribute from SSSD cache if user has no secondary
         groups anymore
        :id: af7fd9fd-e044-461c-ad51-c91c0b371018
        :steps:
          1. Run id <ad user>
          2. Run ldbsearch -H <Domain-cache> name=AD user
          3. Remove AD users membership of Group
          4. Stop, clear cache and start sssd
          5. Run ldbsearch -H <Domain-cache> name=AD user
        :expectedresults:
          1. AD Users cache Entry should not have OriginalMember of
             attribute having Windows AD Group DN
        """
        adjoin(membersw='adcli')
        (aduser, adgroup) = create_aduser_group
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        basedn_entry = multihost.ad[0].domain_basedn_entry
        users_dn_entry = f'CN=Users,{basedn_entry}'
        ad_group_dn = f'CN={adgroup},{users_dn_entry}'
        domain = multihost.ad[0].domainname
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        client.clear_sssd_cache()
        user_id = f'id {aduser}@{domain}'
        multihost.client[0].run_command(user_id)
        user_cache_entry = 'name=%s@%s,cn=users'\
                           ',cn=%s,cn=sysdb' % (aduser, domain.lower(), domain)
        ldb_search = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb  -b ' \
                     '%s originalMemberOf' % (domain_name, user_cache_entry)
        cmd = multihost.client[0].run_command(ldb_search)
        results = cmd.stdout_text.split()
        # Remove user from group and clear sssd cache
        if ad_group_dn in results:
            client.remove_ad_user_group(aduser)
            client.remove_ad_user_group(adgroup)
            client.clear_sssd_cache()
            multihost.client[0].run_command(user_id, raiseonerr=False)
            cmd = multihost.client[0].run_command(ldb_search, raiseonerr=False)
            results = cmd.stdout_text.split()
            assert ad_group_dn not in results
        else:
            pytest.fail("%s cmd failed" % ldb_search)

    @pytest.mark.tier1
    def test_0002_bz1287209(self, multihost, adjoin, create_aduser_group):
        """
        :title: IDM-SSSD-TC: ad_provider: ad_parameters: Allow short usernames
         in trust setups
        :id: 651a0fb6-7199-40af-aafb-5edff3e17d39
        :customerscenario: True
        :steps:
          1. Modify sssd.conf and set "full_name_format=%1$s" in  Domain
             section
          2. Run command "su -ADUser@Domain -c whoami"
        :expectedresults:
          1. Should succeed
          2. Output of whoami command should display only the User part without
             Domain part ,Ex: "ADUser"
        """
        adjoin(membersw='adcli')
        (ad_user, _) = create_aduser_group
        client_ad = sssdTools(multihost.client[0], multihost.ad[0])
        domain = multihost.ad[0].domainname
        client_ad.sssd_conf(f'domain/{domain}', {'full_name_format': '%1$s'})
        client_ad.clear_sssd_cache()
        su_cmd = f'su - {ad_user}@{domain} -c  whoami'
        cmd = multihost.client[0].run_command(su_cmd, raiseonerr=False)
        assert ad_user == cmd.stdout_text.strip()

    @pytest.mark.tier1
    def test_0003_bz1421622(self, multihost, adjoin, create_aduser_group):
        """
        :title: IDM-SSSD-TC: ad_provider: ad_parameters: Users or Groups are
         cached as mixed-case resulting in users unable to sign in
        :id: 164201c2-61f4-4bbc-b936-fd0050d2fa08
        :steps:
          1. Run command "#getent group <group_name>"
          2. Run command "#sssctl group-show <group_name>"
        :expectedresults:
          1. Group look up should successful.
          2. Get successful information about cached group.
        """
        logger_cmd = 'logger test_0003_bz1421622'
        multihost.client[0].run_command(logger_cmd, raiseonerr=False)
        adjoin(membersw='adcli')
        (_, _) = create_aduser_group
        domain = multihost.ad[0].domainname.strip().upper()
        client = sssdTools(multihost.client[0], multihost.ad[0])
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        client.clear_sssd_cache()
        userlist = ['users', 'Users', 'USERS', 'uSERS', 'UsErS', 'uSeRs',
                    'users']
        domainlist = ['domain', 'Domain', 'DOMAIN', 'dOMAIN', 'DoMaIn',
                      'dOmAiN', 'DOMAIN']
        for i in zip(userlist, domainlist):
            grp = (i[1] + ' ' + i[0] + "@" + domain)
            cmd1 = multihost.client[0].run_command(['getent', 'group', grp],
                                                   raiseonerr=False)
            cmd2 = multihost.client[0].run_command(['sssctl', 'group-show',
                                                    grp], raiseonerr=False)
            if cmd1.returncode == 0 and cmd2.returncode == 0:
                assert True
            else:
                assert False

    @pytest.mark.tier1
    def test_00015_authselect_cannot_validate_its_own_files(self, multihost, adjoin):
        """
        :title: authselect: authselect cannot validate its own files
        :id: 67bec814-d67b-4469-9662-58354889d549
        :requirement: IDM-SSSD-REQ :: Authselect replaced authconfig
        :casecomponent: authselect
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1734302
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = multihost.ad[0].domainname.strip().upper()
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        client.clear_sssd_cache()
        multihost.client[0].run_command("yum install -y gdb")
        multihost.client[0].run_command("gdb -quiet authselect -ex "
                                        "'set breakpoint pending on' -ex "
                                        "'b src/lib/files/system.c:428' -ex "
                                        "'run select sssd --force' -ex "
                                        "'shell sleep 1' -ex 'detach' -ex "
                                        "'quit'")
        cmd_check = multihost.client[0].run_command("authselect check")
        assert "Current configuration is valid." in cmd_check.stdout_text

    @pytest.mark.tier1
    def test_0005_BZ1527149_BZ1549675(self, multihost, adjoin, create_adgrp):
        """
        :title: IDM-SSSD-TC: ad_provider: ad_parameters: AD BUILTIN groups are
         cached with gidNumber equal to 0
        :id: d31bffa6-4313-44af-b103-9ea4bc715e72
        :customerscenario: True
        :steps:
          1. Create AD group with scope as "Global" and type "Security"
          2. Update the properties newly created group and update under
             "Member of" tab and add Users BUILTIN group.
          3. Check the group lookup for BUILTIN group.
          4. Check the cache entry, for built in group.
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Group lookup should give empty output.
          4. Should not list the entry for built in group
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        domain = multihost.ad[0].domainname.strip()
        user = f'Administrator@{domain}'
        ldbcache = f'/var/lib/sss/db/cache_{domain_name}.ldb'
        user_cache_entry = f'name={user},cn=group,cn={domain},cn=sysdb'
        # just to check sssd status checking following user lookup
        getent_pwd_cmd = f"getent passwd {user}"
        cmd = multihost.client[0].run_command(getent_pwd_cmd, raiseonerr=False)
        if cmd.returncode == 0:
            getent_cmd = f"getent group {user}"
            cmd = multihost.client[0].run_command(getent_cmd, raiseonerr=False)
            if cmd.returncode != 0:
                ldbcmd = f"ldbsearch -H {ldbcache} -b {user_cache_entry}"
                cmd = multihost.client[0].run_command(ldbcmd, raiseonerr=False)
                if cmd.returncode == 0:
                    ldb_search_entry = cmd.stdout_text.strip().split('\n')[0]
                    assert '# returned 0 records' in ldb_search_entry
            else:
                pytest.fail("Expected to get empty output for group lookup")

    @pytest.mark.tier1
    def test_0006_bz1592964(self, multihost, adjoin, create_aduser_group,
                            create_domain_local_group,
                            add_user_in_domain_local_group):
        """
        :title: IDM-SSSD-TC: ad_provider: ad_parameters: Groups go missing
         with PAC enabled in sssd
        :id: 505be110-0b3c-46ea-8be8-15c9ee2291f4
        :customerscenario: True
        :steps:
          1. Update sssd with pac.
          2. Remove the sssd cache.
          3. Check user lookup and check entry for all groups in output.
          4. Check ldbsearch and check entry for all groups for that user
        :expectedResults
          1. Successfully update PAC enabled service in sssd.
          2. Successfully remove the sssd cache.
          3. User and group lookup should be successfull
          4. Get successful information about domain local group in ldbsearch.
        """
        adjoin(membersw='adcli')
        (ad_user, _) = create_aduser_group
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        cfgget = '/etc/sssd/sssd.conf'
        sssdcfg = multihost.client[0].get_file_contents(cfgget)
        sssdcfg = sssdcfg.replace(b'services = nss, pam',
                                  b'services = nss, pam, pac')
        multihost.client[0].put_file_contents(cfgget, sssdcfg)
        client.fix_sssd_conf_perms()
        multihost.client[0].run_command('sss_cache -E')
        multihost.client[0].service_sssd('restart')
        time.sleep(20)
        domain = multihost.ad[0].domainname.strip().lower()
        id_lookup = 'id %s@%s' % (ad_user, domain)
        cmd1 = multihost.client[0].run_command(id_lookup, raiseonerr=False)
        user_cache_entry = 'name=%s@%s,cn=users,cn=%s,cn=sysdb' % (ad_user,
                                                                   domain,
                                                                   domain)
        ldb_search = 'ldbsearch -H /var/lib/sss/db/cache_%s.ldb -b %s ' \
                     'name=%s* memberof' % (domain_name, user_cache_entry,
                                            ad_user)

        cmd2 = multihost.client[0].run_command(ldb_search, raiseonerr=False)
        grouplist = ['ltestgroup1', 'ltestgroup2', 'ltestgroup3',
                     'ltestgroup4', 'ltestgroup5']
        for _, group in enumerate(grouplist):
            assert group in cmd1.stdout_text and cmd2.stdout_text

    @pytest.mark.tier2
    def test_0007_bz1361597(self, multihost, adjoin, create_aduser_group):
        """
        :title: IDM-SSSD-TC: ad_provider: ad_parameters: Test if we can lookup
         the AD group with one member
        :id: 03cde0b0-27ae-43bd-84ff-96bceb0a15db
        :steps:
          1. Lookup the group.
          2. Add only one user to the group as the member.
          3. Lookup the group.
          4. Delete the member.
          5. Lookup the group.
        :expectedresults:
          1. Look must be successful.
          2. User must be added to the group as a member successfully.
          3. Look must be successful.
          4. Member should be deleted successfully.
          5. Lookup must be successful.
        """
        adjoin(membersw='adcli')
        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'entry_cache_timeout': '30'}
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        digit = random.randint(1000, 10000)
        group = "testgrp_%s" % digit
        user = "testusr_%s" % digit
        ad_realm = multihost.ad[0].domainname
        password = multihost.ad[0].ssh_password
        adcli_cg = f"adcli create-group {group} --domain={ad_realm}"\
            f" --login-user=Administrator --stdin-password -v"
        multihost.client[0].run_command(adcli_cg, stdin_text=password)

        adcli_cu = f"adcli create-user {user} --domain={ad_realm}"\
            f" --login-user=Administrator --stdin-password -v"
        multihost.client[0].run_command(adcli_cu, stdin_text=password)
        time.sleep(30)
        getent = "date; SSS_NSS_USE_MEMCACHE=NO "\
                 "getent group %s@%s" % (group, ad_realm)
        lookup = multihost.client[0].run_command(getent, raiseonerr=False)
        assert lookup.returncode == 0
        adcli_am = f"adcli add-member {group} {user} --domain={ad_realm} "\
            f"--login-user=Administrator  --stdin-password -v"
        multihost.client[0].run_command(adcli_am, stdin_text=password)
        time.sleep(30)
        lookup = multihost.client[0].run_command(getent, raiseonerr=False)
        print(lookup.stdout_text)
        adcli_rm = f"adcli remove-member {group} {user} --domain={ad_realm} "\
            f"--login-user=Administrator  --stdin-password -v"
        multihost.client[0].run_command(adcli_rm, stdin_text=password)
        time.sleep(30)
        lookup = multihost.client[0].run_command(getent, raiseonerr=False)
        print(lookup.stdout_text)
        search = "%s@%s" % (user, ad_realm)
        assert lookup.stdout_text.find(search) == -1
        print("Delete user and group")
        adcli_dg = f"adcli delete-group {group} --domain={ad_realm}"\
            f" --login-user=Administrator  --stdin-password -v"
        multihost.client[0].run_command(adcli_dg, stdin_text=password)
        adcli_du = f"adcli delete-user {user} --domain={ad_realm}"\
            f" --login-user=Administrator  --stdin-password -v"
        multihost.client[0].run_command(adcli_du, stdin_text=password)
        print("Group {} and User {} deleted".format(group, user))
        assert lookup.returncode == 0

    @pytest.mark.tier3
    def test_0008_bz1431858(self, multihost, adjoin):
        """
        :title : IDM-SSSD-TC: ad_provider: ad_parameters: Wrong principal is
         found with ad provider having long hostname
        :id: 8ce3e54b-6b4d-4a35-b751-47020db1ac97
        :steps:
          1. Provision windows hostname longer than 15 chracters
          2. Provision RHEL machine and install SSSD
          3. Add a user to AD.
          4. Lookup user
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Lookup should be successful
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = multihost.ad[0].domainname.strip().upper()
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        user = "Administrator"
        ad_realm = multihost.ad[0].domainname
        cmd = "id %s@%s" % (user, ad_realm)
        multihost.client[0].run_command(cmd, raiseonerr=False)
        output = multihost.client[0].run_command('klist -kt').stdout_text
        search = f"host/{multihost.client[0].external_hostname}"
        assert output.find(search) != -1

    @pytest.mark.tier1
    def test_0009_bz1565761(self, multihost, adjoin):
        """
        :title : IDM-SSSD-TC: ad_provider: ad_parameters: SSSD reports minor
         failures trying to resolve well-known SIDs
        :id: 65cc1f42-92b0-4e3e-8752-5ed9bac2fd6d
        :customerscenario: True
        :steps:
          1. Lookup user
          2. Grep "Domain not found"
        :expectedresults:
          1. Lookup should be successful.
          2. Empty output
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = multihost.ad[0].domainname.strip().upper()
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        user = "Administrator"
        client.clear_sssd_cache()
        multihost.client[0].run_command(
            f"id {user}@{domain}", raiseonerr=False
        )
        grep = 'grep -ire "Domain not found" /var/log/sssd/'
        cmd = multihost.client[0].run_command(grep, raiseonerr=False)
        output = cmd.stdout_text
        assert not output  # output should be empty

    @pytest.mark.tier1
    def test_0010_bz1527662(self, multihost, adjoin):
        """
        :title: ad_parameters: Handle conflicting e-mail addresses
         more gracefully
        :id: 21b13b8f-0fc5-44e0-9ce0-e59f74826db0
        :customerscenario: True
        :steps:
          1. create ad user akhomic1 having mail akhomic1b@<domain>
          2. create ad user akhomic1b
          3. login as akhomic1 user
        :expectedresults:
          1. akhomic1 and akhomic1b should  successfully login
        """
        adjoin(membersw='adcli')
        user_list = ['akhomic1', 'akhomic1b']
        ad_realm = multihost.ad[0].domainname
        user_mail = 'akhomic1b@%s' % ad_realm
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = multihost.ad[0].domainname.strip().upper()
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        for user in user_list:
            group = '%s_group' % (user)
            client.create_ad_user(user, group, user_mail)
        multihost.client[0].service_sssd('restart')
        result = True
        for user in user_list:
            ad_user = '%s@%s' % (user, ad_realm)
            res = client.auth_from_client(ad_user, 'Secret123')
            result = result and (res == 3)

        for user in user_list:
            group = '%s_group' % (user)
            client.remove_ad_user_group(group)
            client.remove_ad_user_group(user)
        assert result, "One ore more users failed to login!"

    @pytest.mark.tier1
    def test_0011_bz1571526(self, multihost, adjoin):
        """
        :title: ad_parameters: sssd should give warning
         when changing ldap schema from AD to others
        :id: ae2fc714-e698-492f-8d80-4d180e049cc3
        :customerscenario: True
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'ldap_schema': 'rfc2307', 'debug_level': '9'}
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        domain_log = '/var/log/sssd/sssd_%s.log' % domain_name
        log = multihost.client[0].get_file_contents(domain_log).decode('utf-8')
        msg = 'The AD provider only supports the AD LDAP schema. '\
              'SSSD will ignore the ldap_schema option value and '\
              'proceed with ldap_schema=ad'
        find = re.compile(r'%s' % msg)
        assert find.search(log)

    @pytest.mark.tier1
    def test_0012_bz1738532(self, multihost, adjoin, create_aduser_group):
        """
        :title: ad_parameters: lookup identity does not work in some cases
        :id: b8382774-e568-4e5b-b787-bdd4db380c28
        :steps:
          1. Add user and set its UPN different from the username,
             Ex: TestUserUPN@ad.vm
          2. Run command "dbus-send --print-reply --system
             --dest=org.freedesktop.sssd.infopipe /org/freedesktop/sssd/
             infopipe org.freedesktop.sssd.infopipe.GetUserAttr string:
             TestUserUPN@ad.vm array:string:name"
        :expectedresults:
          1. Should succeed
          2. Output of above command should not give any error message
             suppose to get username in output.
        """
        adjoin(membersw='adcli')
        (ad_user, _) = create_aduser_group
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain = multihost.ad[0].domainname.strip().lower()
        user = f'{ad_user}@{domain}'
        client.sssd_conf(f'domain/{domain}', {'debug_level': '9'})
        client.clear_sssd_cache()
        set_UPN = 'powershell.exe -inputformat none -noprofile Set-ADUser ' \
                  '-UserPrincipalName TestUserUPN@ad.vm -Identity %s' % ad_user
        cmd = multihost.ad[0].run_command(set_UPN, raiseonerr=False)
        if cmd.returncode != 0:
            print('Failed to set UPN of user')
            status = 'FAIL'
            test_cmd = 'dbus-send --print-reply --system --dest=org.' \
                       'freedesktop.sssd.infopipe /org/freedesktop/sssd/' \
                       'infopipe org.freedesktop.sssd.infopipe.' \
                       'GetUserAttrstring:TestUserUPN@ad.vm array:string:name'
            check = multihost.client[0].run_command(test_cmd, raiseonerr=False)
            if check.returncode == 0:
                status = 'PASS'
                find = re.compile(r'%s' % user)
                result = find.search(check.stdout_text)
                if result is None:
                    status = 'FAIL'
            assert status != 'FAIL'

    @pytest.mark.tier1
    def test_0013_bz1794016(self, multihost, adjoin):
        """
        :title: sssd_be frequently crashes when refresh_expired_interval
         is set and multiple domains are configured
        :id: ac5e99cc-2d81-48f7-beef-622f1c6b3684
        """
        adjoin(membersw='adcli')
        client = sssdTools(multihost.client[0])
        client.backup_sssd_conf()

        domain_sec_name = client.get_domain_section_name()
        sssd_params = {'domains': f'local, {domain_sec_name}'}
        client.sssd_conf('sssd', sssd_params)
        domain_params = {'entry_cache_timeout': '5400',
                         'refresh_expired_interval': '4000'
                         }
        client.sssd_conf(f'domain/{domain_sec_name}', domain_params)
        local_params = {'proxy_lib_name': 'files',
                        'proxy_pam_target': 'sssd-shadowutils',
                        'id_provider': 'proxy'}
        client.sssd_conf('domain/local', local_params)
        client.clear_sssd_cache()
        journalctl = 'journalctl -x -n 150 --no-pager'
        multihost.client[0].service_sssd('restart')
        journal_output = multihost.client[0].run_command(journalctl)
        coredump = re.compile(r'%s' % '(sssd_be) of user 0 dumped core.')
        result = coredump.search(journal_output.stdout_text)
        client.restore_sssd_conf()
        assert result is None

    @pytest.mark.tier1
    def test_0014_user_filtering(self, multihost,
                                 adjoin, create_aduser_group):
        """
        :title: SSSD user filtering is failing
         after files provider rebuilds cache
        :id: a85acb65-a8af-4397-b4e0-fa9e093b86d7
        :customerscenario: True
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1824323
        :steps:
          1. Join RHEL to the AD-server
          2. Create two ADusers on AD
          3. Add one ADuser in filter_users in nss section
          4. Enable files domain in sssd.conf
          5. Restart sssd with cleared cache
          6. Make sure that filtered user is not returned by SSSD
          7. Add a local user and fetch that user information
          8. Again Make sure that filtered user is not returned by SSSD
        :expectedresults:
          1. Should Succeed
          2. Should Succeed
          3. Should Succeed
          4. Should Succeed
          5. Should Succeed
          6. filtered user should not be returned
          7. filtered user should not be returned after localuser addition
          8. Other AD-users should be returned correctly
        """
        if not multihost.client[0].detect_files_provider():
            pytest.skip("Files Provider support isn't available, skipping")
        adjoin(membersw='adcli')
        (aduser1, _) = create_aduser_group
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domainname = multihost.ad[0].domainname
        params1 = {'enable_files_domain': 'true'}
        client.sssd_conf('sssd', params1)
        client.clear_sssd_cache()
        aduser2 = f'administrator@{domainname}'
        lkup1 = f'getent passwd -s sss {aduser1}@{domainname}'
        multihost.client[0].run_command(lkup1, raiseonerr=True)
        lkup2 = f'getent passwd -s sss {aduser2}'
        multihost.client[0].run_command(lkup2, raiseonerr=True)
        multihost.client[0].service_sssd('stop')
        userlist = f'root, {aduser2}'
        params2 = {'filter_users': userlist,
                   'filter_groups': 'root'}
        client.sssd_conf('nss', params2)
        client.remove_sss_cache('/var/lib/sss/db')
        client.remove_sss_cache('/var/lib/sss/mc')
        client.remove_sss_cache('/var/log/sssd')
        multihost.client[0].service_sssd('start')
        multihost.client[0].run_command(lkup1, raiseonerr=True)
        cmd = multihost.client[0].run_command(lkup2, raiseonerr=False)
        assert cmd.returncode == 2
        usradd = '/usr/sbin/useradd localuser-test'
        multihost.client[0].run_command(usradd, raiseonerr=True)
        lkup3 = 'getent passwd localuser-test'
        multihost.client[0].run_command(lkup3, raiseonerr=True)
        multihost.client[0].run_command(lkup1, raiseonerr=True)
        cmd = multihost.client[0].run_command(lkup2, raiseonerr=False)
        usrdel = '/usr/sbin/userdel -rf localuser-test'
        multihost.client[0].run_command(usrdel, raiseonerr=True)
        multihost.client[0].service_sssd('stop')
        client.sssd_conf('sssd', params1, action='delete')
        client.sssd_conf('nss', params2, action='delete')
        multihost.client[0].service_sssd('start')
        assert cmd.returncode == 2

    @pytest.mark.tier1
    def test_0016_forceLDAPS(self, multihost,
                             adjoin, fetch_ca_cert,
                             create_aduser_group):
        """
        :title: Force LDAPS over 636 with AD Access Provider
        :id: 12d0c340-9c50-4583-97c1-23f9f583522c
        :customerscenario: True
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1762415
        :steps:
          1. Join RHEL to the AD-server
          2. Block 389 port on client with iptable
          3. Enable 'ad_use_ldaps' option sssd.conf
          4. Add channel bindings /etc/openldap/ldap.conf
          4. Restart sssd
          5. Run id <Username>
          6. Parse sssd log file for port used to contact AD
        :expectedresults:
          1. User information should be returned correctly
          2. Logs should show that port 636 was used to contact AD
        """
        winver = multihost.ad[0].run_command(
            "systeminfo", raiseonerr=False).stdout_text
        if "Microsoft Windows Server 2012" in winver:
            pytest.skip("Test not valid on windows 2012R2.")
        adjoin(membersw='adcli')
        (aduser, adgroup) = create_aduser_group
        client = sssdTools(multihost.client[0], multihost.ad[0])
        cert_bkup = 'cp /etc/openldap/ldap.conf /etc/openldap.conf_bk'
        cmd = multihost.client[0].run_command(cert_bkup, raiseonerr=False)
        cert_conf = 'echo -e "SASL_CBINDING tls-endpoint\n"\
                    "TLS_CACERT /etc/openldap/certs/cacert.pem\n"\
                    "SASL_NOCANON	on" > /etc/openldap/ldap.conf'
        cmd = multihost.client[0].run_command(cert_conf, raiseonerr=False)
        cat = 'cat /etc/openldap/ldap.conf'
        cmd = multihost.client[0].run_command(cat, raiseonerr=False)
        domainname = multihost.ad[0].domainname
        lkup = 'getent passwd %s@%s' % (aduser, domainname)
        cmd4 = multihost.client[0].run_command(lkup, raiseonerr=False)
        multihost.client[0].service_sssd('stop')
        cmd = 'dnf install -y firewalld'
        multihost.client[0].run_command(cmd, raiseonerr=True)
        cmd = 'systemctl start firewalld'
        multihost.client[0].run_command(cmd, raiseonerr=True)
        fw_add1 = 'firewall-cmd --permanent --direct --add-rule ipv4 '\
                  'filter OUTPUT 0 -p tcp -m tcp --dport=389 -j DROP'
        fw_add2 = 'firewall-cmd --permanent --direct --add-rule ipv4 '\
                  'filter OUTPUT 1 -j ACCEPT'
        multihost.client[0].run_command(fw_add1, raiseonerr=True)
        multihost.client[0].run_command(fw_add2, raiseonerr=True)
        fw_rld = 'firewall-cmd --reload'
        multihost.client[0].run_command(fw_rld, raiseonerr=True)
        domain_section = 'domain/{}'.format(domainname)
        sssd_params = {'ad_use_ldaps': 'True',
                       'ldap_id_mapping': 'True',
                       'debug_level': '9'}
        client.sssd_conf(domain_section, sssd_params)
        client.remove_sss_cache('/var/lib/sss/db')
        client.remove_sss_cache('/var/log/sssd')
        multihost.client[0].service_sssd('start')
        time.sleep(3)
        lkup = 'getent passwd %s@%s' % (aduser, domainname)
        cmd4 = multihost.client[0].run_command(lkup, raiseonerr=False)
        fw_r1 = 'firewall-cmd --permanent --direct --remove-rule ipv4 '\
                'filter OUTPUT 0 -p tcp -m tcp --dport=389 -j DROP'
        fw_r2 = 'firewall-cmd --permanent --direct --remove-rule ipv4 '\
                'filter OUTPUT 1 -j ACCEPT'
        multihost.client[0].run_command(fw_r1, raiseonerr=True)
        multihost.client[0].run_command(fw_r2, raiseonerr=True)
        multihost.client[0].run_command(fw_rld, raiseonerr=True)
        cmd = 'systemctl stop firewalld'
        multihost.client[0].run_command(cmd, raiseonerr=True)
        cmd = 'dnf remove -y firewalld'
        multihost.client[0].run_command(cmd, raiseonerr=True)
        cert_restr = 'mv /etc/openldap.conf_bk /etc/openldap/ldap.conf'
        cmd = multihost.client[0].run_command(cert_restr, raiseonerr=False)
        multihost.client[0].service_sssd('stop')
        client.sssd_conf(domain_section, sssd_params, action='delete')
        multihost.client[0].service_sssd('start')
        assert cmd4.returncode == 0

    @pytest.mark.tier2
    def test_0017_gssspnego_adjoin(self, multihost):
        """
        :title: Verify sssd uses GSS-SPNEGO when communicating to AD
        :id: 9d8b68a0-1208-446c-9dbd-93ee1f934903
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1707963
        """
        ad_hostname = multihost.ad[0].sys_hostname
        status = ''
        pcapfile = '/tmp/spnego.pcap'
        tcpdump_cmd = 'tcpdump -s0 host %s -w %s' % (ad_hostname, pcapfile)
        ad_domain = multihost.ad[0].domainname
        realm_join_cmd = 'realm join %s --client-software=sssd '\
                         '--server-software=active-directory '\
                         '--membership-software=adcli -v' % (ad_domain)
        multihost.client[0].run_command(tcpdump_cmd, bg=True)
        pkill = 'pkill tcpdump'
        cmd = multihost.client[0].run_command(realm_join_cmd,
                                              stdin_text='Secret123',
                                              raiseonerr='False')
        if cmd.returncode != 0:
            status = 'FAIL'
            print("Joining to %s failed" % ad_domain)
        multihost.client[0].run_command(pkill, raiseonerr=False)
        tshark_cmd = "tshark -r %s -V -2 -R " \
                     "'ldap.mechanism == GSS-SPNEGO'" % pcapfile
        valid_etypes = 'etype: eTYPE-AES256-CTS-HMAC-SHA1-96'
        check_str = re.compile(r'%s' % valid_etypes)
        cmd = multihost.client[0].run_command(tshark_cmd, raiseonerr=False)
        if cmd.returncode != 0:
            status = 'FAIL'
            print("%s failed " % tshark_cmd)
        else:
            if not check_str.search(cmd.stdout_text):
                status = 'FAIL'
        bindResponse = "tshark -r %s -V -2 -R "\
                       "'spnego.krb5.tok_id == 0x0002'" % pcapfile
        cmd = multihost.client[0].run_command(bindResponse, raiseonerr=False)
        if not check_str.search(cmd.stdout_text):
            status = 'FAIL'
        else:
            status = 'PASS'
        realm_leave = "realm leave -v %s" % (ad_domain)
        multihost.client[0].run_command(realm_leave)
        remove_pcap = 'rm -f %s' % pcapfile
        multihost.client[0].run_command(remove_pcap)
        assert status == 'PASS'



"""Automation for default debug level

:requirement: SSSD - Default debug level
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

from __future__ import print_function
import re
import time
import pytest
from sssd.testlib.common.utils import sssdTools, LdapOperations
from constants import ds_instance_name, ds_suffix, ds_rootpw, ds_rootdn


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.defaultdebuglevel
class TestDefaultDebugLevel(object):
    """ Check sssd default debug level """
    @staticmethod
    @pytest.mark.tier2
    def test_bz785908(multihost, backupsssdconf):
        """
        :title: ldap search base dose not fully limit the group search base bz785908
        :id: 5f736f56-a299-11ed-a880-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=785908
        :setup:
            1. Add users to Ldap server
            2. Create group with member within group search base
            2. Create group with member outside of the group search base
            4. Update sssd domain with ldap_group_object_class, ldap_search_base and
                ldap_group_object_class
        :steps:
            1. Runs command getent -s sss group to query the SSSD cache and save the results to a file.
            2. Cleans up the LDAP server by deleting the previously created user and group entries,
               and finally search for log messages containing the strings "Group111" and "Group22".
        :expectedresults:
            1. Query the SSSD cache and save the results to a file Should succeed
            2. Log messages should containing the strings
                "Group111" and "Group22" in the /tmp/grp_file
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        user_dn = f'uid=tempuser2,{ds_suffix}'
        user_info = {'cn': 'Temp user2'.encode('utf-8'),
                     'loginshell': '/bin/bash'.encode('utf-8'),
                     'uidNumber': '121298'.encode('utf-8'),
                     'gidNumber': '10000'.encode('utf-8'),
                     'objectClass': ['top'.encode('utf-8'),
                                     'posixAccount'.encode('utf-8')],
                     'homeDirectory': '/home/tempuser2'.encode('utf-8'),
                     'userPassword': 'Secret123'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        user_dn = f'uid=tempuser3,{ds_suffix}'
        user_info = {'cn': 'Temp user3'.encode('utf-8'),
                     'loginshell': '/bin/bash'.encode('utf-8'),
                     'uidNumber': '121297'.encode('utf-8'),
                     'gidNumber': '10000'.encode('utf-8'),
                     'objectClass': ['top'.encode('utf-8'),
                                     'posixAccount'.encode('utf-8')],
                     'homeDirectory': '/home/tempuser3'.encode('utf-8'),
                     'userPassword': 'Secret123'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        ldap_inst.org_unit("qagroup", ds_suffix)
        group_dn = f"cn=Group11,ou=qagroup,{ds_suffix}"
        grp_info = {'cn': 'Group11'.encode('utf-8'),
                    'gidNumber': '222011'.encode('utf-8'),
                    'objectClass': ['top'.encode('utf-8'),
                                    'posixGroup'.encode('utf-8'),
                                    'groupOfNames'.encode('utf-8')],
                    'member': f'uid=tempuser3,{ds_suffix}'.encode('utf-8')}
        ldap_inst.add_entry(grp_info, group_dn)
        group_dn = f"cn=Group22,ou=Groups,{ds_suffix}"
        grp_info = {'cn': 'Group22'.encode('utf-8'),
                    'gidNumber': '222022'.encode('utf-8'),
                    'objectClass': ['top'.encode('utf-8'),
                                    'posixGroup'.encode('utf-8'),
                                    'groupOfNames'.encode('utf-8')],
                    'member': f'uid=tempuser3,{ds_suffix}'.encode('utf-8')}
        ldap_inst.add_entry(grp_info, group_dn)
        group_dn = f"cn=Group222,ou=qagroup,{ds_suffix}"
        grp_info = {'cn': 'Group222'.encode('utf-8'),
                    'gidNumber': '222000'.encode('utf-8'),
                    'objectClass': ['top'.encode('utf-8'),
                                    'posixGroup'.encode('utf-8'),
                                    'groupOfNames'.encode('utf-8')],
                    'member': [f'cn=Group11,ou=qagroup,{ds_suffix}'.encode('utf-8')]}
        ldap_inst.add_entry(grp_info, group_dn)
        group_dn = f"cn=Group111,ou=Groups,{ds_suffix}"
        grp_info = {'cn': 'Group111'.encode('utf-8'),
                    'gidNumber': '111000'.encode('utf-8'),
                    'objectClass': ['top'.encode('utf-8'),
                                    'posixGroup'.encode('utf-8'),
                                    'groupOfNames'.encode('utf-8')],
                    'member': [f'cn=Group222,ou=qagroup,{ds_suffix}'.encode('utf-8'),
                               f'cn=Group22,ou=Groups,{ds_suffix}'.encode('utf-8'),
                               f'uid=tempuser2,{ds_suffix}'.encode('utf-8')]}
        ldap_inst.add_entry(grp_info, group_dn)
        tools.sssd_conf("domain/example1",
                        {'cache_credentials': True,
                         'enumerate': True,
                         'ldap_search_base': f'ou=Groups,{ds_suffix}',
                         'ldap_schema': 'rfc2307bis',
                         'ldap_group_object_class': "groupOfNames"}, action='update')
        tools.clear_sssd_cache()
        time.sleep(20)
        cmd = client.run_command("getent -s sss group").stdout_text
        for dn_dn in [f'uid=tempuser2,{ds_suffix}',
                      f'uid=tempuser3,{ds_suffix}',
                      f'cn=Group111,ou=Groups,{ds_suffix}',
                      f'cn=Group222,ou=qagroup,{ds_suffix}',
                      f'cn=Group22,ou=Groups,{ds_suffix}',
                      f'cn=Group11,ou=qagroup,{ds_suffix}',
                      f'ou=qagroup,{ds_suffix}']:
            ldap_inst.del_dn(dn_dn)
        for group in ["Group111", "Group22"]:
            assert group in cmd

    @staticmethod
    @pytest.mark.tier1_4
    def test_bz785898(multihost, backupsssdconf):
        """
        :title: Enable midway cache refresh by default bz785898
        :id: 39695a7a-a29c-11ed-b2b1-845cf3eff344
        :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=785898
        :setup:
            1. Creates a user in the LDAP server for use in subsequent tests.
            3. Updates the SSSD configuration file for the specified domain with some configuration options.
            4. Updates the SSSD configuration file for the "nss" service with some configuration options.
            5. Updates the SSSD configuration file for the "sssd" service with a configuration option.
            6. Clears the SSSD cache.
            7. Creates a directory in the client machine's /tmp directory.
        :steps:
            1. Loops 30 times and performs the following actions in each iteration:
                Truncates the sssd_example1.log file.
                Calls the getent passwd command to retrieve information about the mid_cacheuser user.
                Waits for 3 seconds.
                Copies the contents of the sssd_example1.log file to a file in the /tmp/lookup directory.
            2. Search the /tmp/lookup directory for log
                entries containing the string "Got request for".
            3. Saves the output of the search to the /tmp/lookupfile file.
            4. Deletes the user created before.
            5. Removes the /tmp/lookup and /tmp/lookupfile directories.
            6. Reads the contents of the /tmp/lookupfile file and asserts that there are more than 2 lines.
        :expectedresults:
            1. All commands executed within the loop should Success
            2. "Got request for" string should present
            3. Successfully saved Output of the search to the /tmp/lookupfile file
            4. User deletion should success
            5. Removal of directories should success
            6. There should be more that 2 lines in /tmp/lookupfile
        """
        client = multihost.client[0]
        tools = sssdTools(multihost.client[0])
        ldap_uri = f'ldap://{multihost.master[0].sys_hostname}'
        ldap_inst = LdapOperations(ldap_uri, ds_rootdn, ds_rootpw)
        user_dn = f'uid=mid_cacheuser,{ds_suffix}'
        user_info = {'cn': 'Midcache'.encode('utf-8'),
                     'loginshell': '/bin/bash'.encode('utf-8'),
                     'uidNumber': '121298'.encode('utf-8'),
                     'gidNumber': '10000'.encode('utf-8'),
                     'objectClass': ['top'.encode('utf-8'),
                                     'posixAccount'.encode('utf-8')],
                     'homeDirectory': '/home/mid_cacheuser'.encode('utf-8'),
                     'userPassword': 'Secret123'.encode('utf-8')}
        ldap_inst.add_entry(user_info, user_dn)
        tools.sssd_conf("domain/example1",
                        {'entry_cache_timeout': '60',
                         'cache_credentials': True,
                         'use_fully_qualified_names': False}, action='update')
        tools.sssd_conf("nss", {'filter_groups': 'root',
                                'filter_users': 'root',
                                'debug_level': '9',
                                'entry_cache_nowait_percentage': '50',
                                'memcache_timeout': '1'}, action='update')
        tools.sssd_conf("sssd", {'sbus_timeout': '30'}, action='update')
        tools.clear_sssd_cache()
        client.run_command("mkdir /tmp/lookup")
        for i in range(30):
            client.run_command("> /var/log/sssd/sssd_example1.log")
            client.run_command("getent passwd -s sss mid_cacheuser")
            time.sleep(3)
            client.run_command(f"cat /var/log/sssd/sssd_example1.log > /tmp/lookup/lookup{i}")
        client.run_command("grep -r 'Got request for' /tmp/lookup/")
        client.run_command("grep -r 'Got request for' /tmp/lookup > /tmp/lookupfile")
        time.sleep(3)
        ldap_inst.del_dn(f"uid=mid_cacheuser,{ds_suffix}")
        log_str = multihost.client[0].get_file_contents("/tmp/lookupfile").decode('utf-8')
        for logs in ['lookupfile', 'lookup']:
            client.run_command(f"rm -vfr /tmp/{logs}")
        assert len(log_str.split('\n')) > 2



#
# integration test for sss_override tool
#
# Copyright (c) 2015 Red Hat, Inc.
# Author: Pavel Reichl  <preichl@redhat.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import stat
import ent
import pwd
import config
import signal
import subprocess
import time
import pytest
import ds_openldap
import ldap_ent
import sssd_id
from util import unindent

try:
    from subprocess import check_output
except ImportError:
    # In Python 2.6, the module subprocess does not have the function
    # check_output. This is a fallback implementation.
    def check_output(*popenargs, **kwargs):
        if 'stdout' in kwargs:
            raise ValueError('stdout argument not allowed, it will be '
                             'overridden.')
        process = subprocess.Popen(stdout=subprocess.PIPE, *popenargs,
                                   **kwargs)
        output, _ = process.communicate()
        retcode = process.poll()
        if retcode:
            cmd = kwargs.get("args")
            if cmd is None:
                cmd = popenargs[0]
            raise subprocess.CalledProcessError(retcode, cmd, output=output)
        return output


@pytest.fixture(scope="module")
def ds_inst(request):
    """LDAP server instance fixture"""
    ds_inst = ds_openldap.DSOpenLDAP(
        config.PREFIX, 10389, 'dc=example,dc=com',
        "cn=admin", "Secret123")
    try:
        ds_inst.setup()
    except Exception:
        ds_inst.teardown()
        raise
    request.addfinalizer(lambda: ds_inst.teardown())
    return ds_inst


@pytest.fixture(scope="module")
def ldap_conn(request, ds_inst):
    """LDAP server connection fixture"""
    ldap_conn = ds_inst.bind()
    ldap_conn.ds_inst = ds_inst
    request.addfinalizer(lambda: ldap_conn.unbind_s())
    return ldap_conn


def create_ldap_fixture(request, ldap_conn, ent_list):
    """Add LDAP entries and add teardown for removing them"""
    for entry in ent_list:
        ldap_conn.add_s(entry[0], entry[1])

    def teardown():
        for entry in ent_list:
            ldap_conn.delete_s(entry[0])
    request.addfinalizer(teardown)


def create_conf_fixture(request, contents):
    """Generate sssd.conf and add teardown for removing it"""
    conf = open(config.CONF_PATH, "w")
    conf.write(contents)
    conf.close()
    os.chmod(config.CONF_PATH, stat.S_IRUSR | stat.S_IWUSR)
    request.addfinalizer(lambda: os.unlink(config.CONF_PATH))


def stop_sssd():
    pid_file = open(config.PIDFILE_PATH, "r")
    pid = int(pid_file.read())
    os.kill(pid, signal.SIGTERM)
    while True:
        try:
            os.kill(pid, signal.SIGCONT)
        except OSError:
            break
        time.sleep(1)


def start_sssd():
    """Start sssd"""
    if subprocess.call(["sssd", "-D", "--logger=files"]) != 0:
        raise Exception("sssd start failed")


def restart_sssd():
    stop_sssd()
    start_sssd()


def create_sssd_fixture(request):
    """Start sssd and add teardown for stopping it and removing state"""
    if subprocess.call(["sssd", "-D", "--logger=files"]) != 0:
        raise Exception("sssd start failed")

    def teardown():
        try:
            stop_sssd()
        except Exception:
            pass
        for path in os.listdir(config.DB_PATH):
            os.unlink(config.DB_PATH + "/" + path)
        for path in os.listdir(config.MCACHE_PATH):
            os.unlink(config.MCACHE_PATH + "/" + path)
    request.addfinalizer(teardown)


OVERRIDE_FILENAME = "export_file"


def prepare_sssd(request, ldap_conn, use_fully_qualified_names=False,
                 case_sensitive=True, override_homedir_option=False):
    """Prepare SSSD with defaults"""
    conf_override_homedir_option = ""
    if override_homedir_option:
        conf_override_homedir_option = "override_homedir = /home/ov_option/%u"

    conf = unindent("""\
        [sssd]
        domains             = LDAP
        services            = nss

        [nss]
        memcache_timeout = 1
        {conf_override_homedir_option}

        [domain/LDAP]
        ldap_auth_disable_tls_never_use_in_production = true
        ldap_id_use_start_tls = false
        ldap_schema         = rfc2307
        id_provider         = ldap
        auth_provider       = ldap
        sudo_provider       = ldap
        ldap_uri            = {ldap_conn.ds_inst.ldap_url}
        ldap_search_base    = {ldap_conn.ds_inst.base_dn}
        use_fully_qualified_names = {use_fully_qualified_names}
        case_sensitive      = {case_sensitive}
    """).format(**locals())
    create_conf_fixture(request, conf)
    create_sssd_fixture(request)

    def teardown():
        # remove user export file
        try:
            os.unlink(OVERRIDE_FILENAME)
        except OSError:
            pass
    request.addfinalizer(teardown)


#
# Common asserts for users
#

def assert_user_default():

    # Assert entries are not overriden
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user1')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user1@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user2')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user2@LDAP')

    user1 = dict(name='user1', passwd='*', uid=10001, gid=20001,
                 gecos='User Number 1',
                 dir='/home/user1',
                 shell='/bin/user1_shell')
    user2 = dict(name='user2', passwd='*', uid=10002, gid=20001,
                 gecos='User Number 2',
                 dir='/home/user2',
                 shell='/bin/user2_shell')

    ent.assert_passwd_by_name('user1', user1)
    ent.assert_passwd_by_name('user1@LDAP', user1)

    ent.assert_passwd_by_name('user2', user2)
    ent.assert_passwd_by_name('user2@LDAP', user2)


def assert_user_overriden(override_name=True):

    if override_name:
        name1 = "ov_user1"
        name2 = "ov_user2"
    else:
        name1 = "user1"
        name2 = "user2"

    user1 = dict(name=name1, passwd='*', uid=10010, gid=20010,
                 gecos='Overriden User 1',
                 dir='/home/ov/user1',
                 shell='/bin/ov_user1_shell')

    user2 = dict(name=name2, passwd='*', uid=10020, gid=20020,
                 gecos='Overriden User 2',
                 dir='/home/ov/user2',
                 shell='/bin/ov_user2_shell')

    ent.assert_passwd_by_name('user1', user1)
    ent.assert_passwd_by_name('user1@LDAP', user1)

    if override_name:
        ent.assert_passwd_by_name('ov_user1', user1)
        ent.assert_passwd_by_name('ov_user1@LDAP', user1)

    ent.assert_passwd_by_name('user2', user2)
    ent.assert_passwd_by_name('user2@LDAP', user2)

    if override_name:
        ent.assert_passwd_by_name('ov_user2', user2)
        ent.assert_passwd_by_name('ov_user2@LDAP', user2)


#
# Common fixtures for users
#


@pytest.fixture
def env_two_users_and_group(request, ldap_conn):

    prepare_sssd(request, ldap_conn)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001,
                      gecos='User Number 1',
                      loginShell='/bin/user1_shell',
                      homeDirectory='/home/user1')

    ent_list.add_user("user2", 10002, 20001,
                      gecos='User Number 2',
                      loginShell='/bin/user2_shell',
                      homeDirectory='/home/user2')

    ent_list.add_group("group", 2001,
                       ["user2", "user1"])

    create_ldap_fixture(request, ldap_conn, ent_list)

    # Assert entries are not overriden
    assert_user_default()


@pytest.fixture
def env_two_users_and_group_overriden(request, ldap_conn,
                                      env_two_users_and_group):

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-u", "10010",
                           "-g", "20010",
                           "-n", "ov_user1",
                           "-c", "Overriden User 1",
                           "-h", "/home/ov/user1",
                           "-s", "/bin/ov_user1_shell"])

    subprocess.check_call(["sss_override", "user-add", "user2@LDAP",
                           "-u", "10020",
                           "-g", "20020",
                           "-n", "ov_user2",
                           "-c", "Overriden User 2",
                           "-h", "/home/ov/user2",
                           "-s", "/bin/ov_user2_shell"])

    # Restart SSSD so the override might take effect
    restart_sssd()

    # Assert entries are overriden
    assert_user_overriden()


#
# Simple user override
#


@pytest.fixture
def env_simple_user_override(request, ldap_conn, env_two_users_and_group):

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-u", "10010",
                           "-g", "20010",
                           "-n", "ov_user1",
                           "-c", "Overriden User 1",
                           "-h", "/home/ov/user1",
                           "-s", "/bin/ov_user1_shell"])

    subprocess.check_call(["sss_override", "user-add", "user2@LDAP",
                           "-u", "10020",
                           "-g", "20020",
                           "-n", "ov_user2",
                           "-c", "Overriden User 2",
                           "-h", "/home/ov/user2",
                           "-s", "/bin/ov_user2_shell"])

    # Restart SSSD so the override might take effect
    restart_sssd()


def test_simple_user_override(ldap_conn, env_simple_user_override):
    """Test entries are overriden"""

    assert_user_overriden()


#
# Root user override
#


@pytest.fixture
def env_root_user_override(request, ldap_conn, env_two_users_and_group):

    # Assert entries are not overriden
    ent.assert_passwd_by_name(
        'root',
        dict(name='root', uid=0, gid=0))

    ent.assert_passwd_by_uid(0, dict(name="root"))

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-u", "0",
                           "-g", "0",
                           "-n", "ov_user1",
                           "-c", "Overriden User 1",
                           "-h", "/home/ov/user1",
                           "-s", "/bin/ov_user1_shell"])

    subprocess.check_call(["sss_override", "user-add", "user2",
                           "-u", "10020",
                           "-g", "20020",
                           "-n", "root",
                           "-c", "Overriden User 2",
                           "-h", "/home/ov/user2",
                           "-s", "/bin/ov_user2_shell"])

    # Restart SSSD so the override might take effect
    restart_sssd()


def test_root_user_override(ldap_conn, env_root_user_override):
    """Test entries are not overriden to root"""

    # Override does not have to happen completly, trying to set uid or gid
    # to 0 is simply ignored.
    ent.assert_passwd_by_name(
        'ov_user1',
        dict(name='ov_user1', passwd='*', uid=10001, gid=20001,
             gecos='Overriden User 1',
             dir='/home/ov/user1',
             shell='/bin/ov_user1_shell'))

    # We can create override with name root. This test is just for tracking
    # that this particular behavior won't change.
    ent.assert_passwd_by_name(
        'user2',
        dict(name='root', passwd='*', uid=10020, gid=20020,
             gecos='Overriden User 2',
             dir='/home/ov/user2',
             shell='/bin/ov_user2_shell'))

    ent.assert_passwd_by_uid(0, dict(name="root"))


#
# Override replaces previous override
#


@pytest.fixture
def env_replace_user_override(request, ldap_conn):

    prepare_sssd(request, ldap_conn)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001,
                      gecos='User Number 1',
                      loginShell='/bin/user1_shell',
                      homeDirectory='/home/user1')

    create_ldap_fixture(request, ldap_conn, ent_list)

    # Assert entries are not overriden
    ent.assert_passwd_by_name(
        'user1',
        dict(name='user1', passwd='*', uid=10001, gid=20001,
             gecos='User Number 1',
             dir='/home/user1',
             shell='/bin/user1_shell'))

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-u", "10010",
                           "-g", "20010",
                           "-n", "ov_user1",
                           "-c", "Overriden User 1",
                           "-h", "/home/ov/user1",
                           "-s", "/bin/ov_user1_shell"])

    # Restart SSSD so the override might take effect
    restart_sssd()

    # Assert entries are overriden
    ent.assert_passwd_by_name(
        'user1',
        dict(name='ov_user1', passwd='*', uid=10010, gid=20010,
             gecos='Overriden User 1',
             dir='/home/ov/user1',
             shell='/bin/ov_user1_shell'))

    # Override of override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-u", "10100",
                           "-g", "20100",
                           "-n", "ov2_user1",
                           "-c", "Overriden2 User 1",
                           "-h", "/home/ov2/user1",
                           "-s", "/bin/ov2_user1_shell"])

    # Restart SSSD so the override might take effect
    restart_sssd()


def test_replace_user_override(ldap_conn, env_replace_user_override):

    user = dict(name='ov2_user1', passwd='*', uid=10100, gid=20100,
                gecos='Overriden2 User 1',
                dir='/home/ov2/user1',
                shell='/bin/ov2_user1_shell')

    ent.assert_passwd_by_name('ov2_user1', user)
    ent.assert_passwd_by_name('ov2_user1@LDAP', user)

    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user1')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user1@LDAP')


#
# Override removal
#


@pytest.fixture
def env_remove_user_override(request, ldap_conn,
                             env_two_users_and_group_overriden):

    # Drop all overrides
    subprocess.check_call(["sss_override", "user-del", "user1"])
    subprocess.check_call(["sss_override", "user-del", "user2@LDAP"])

    # Avoid hitting memory cache
    time.sleep(2)


def test_remove_user_override(ldap_conn, env_remove_user_override):

    # Test entries are not overriden
    assert_user_default()


#
# Override import/export
#


@pytest.fixture
def env_imp_exp_user_override(request, ldap_conn,
                              env_two_users_and_group_overriden):

    # Export overrides
    subprocess.check_call(["sss_override", "user-export", OVERRIDE_FILENAME])

    # Drop all overrides
    subprocess.check_call(["sss_override", "user-del", "user1"])
    subprocess.check_call(["sss_override", "user-del", "user2@LDAP"])

    # Avoid hitting memory cache
    time.sleep(2)

    # Assert entries are not overridden
    assert_user_default()

    # Import overrides
    subprocess.check_call(["sss_override", "user-import",
                           OVERRIDE_FILENAME])
    restart_sssd()


def test_imp_exp_user_override(ldap_conn, env_imp_exp_user_override):

    assert_user_overriden()


# Regression test for bug 3179


def test_imp_exp_user_override_noname(ldap_conn,
                                      env_two_users_and_group):

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-u", "10010",
                           "-g", "20010",
                           "-c", "Overriden User 1",
                           "-h", "/home/ov/user1",
                           "-s", "/bin/ov_user1_shell"])

    subprocess.check_call(["sss_override", "user-add", "user2@LDAP",
                           "-u", "10020",
                           "-g", "20020",
                           "-c", "Overriden User 2",
                           "-h", "/home/ov/user2",
                           "-s", "/bin/ov_user2_shell"])

    # Restart SSSD so the override might take effect
    restart_sssd()

    # Assert entries are overriden
    assert_user_overriden(override_name=False)

    # Export overrides
    subprocess.check_call(["sss_override", "user-export", OVERRIDE_FILENAME])

    # Drop all overrides
    subprocess.check_call(["sss_override", "user-del", "user1"])
    subprocess.check_call(["sss_override", "user-del", "user2@LDAP"])

    # Avoid hitting memory cache
    time.sleep(2)

    # Assert entries are not overridden
    assert_user_default()

    # Import overrides
    subprocess.check_call(["sss_override", "user-import",
                           OVERRIDE_FILENAME])
    restart_sssd()

    assert_user_overriden(override_name=False)


#
# Override user-show
#


@pytest.fixture
def env_show_user_override(request, ldap_conn,
                           env_two_users_and_group_overriden):
    pass


def test_show_user_override(ldap_conn, env_show_user_override):

    out = check_output(['sss_override', 'user-show', 'user1']).decode('utf-8')
    assert out == "user1@LDAP:ov_user1:10010:20010:Overriden User 1:"\
                  "/home/ov/user1:/bin/ov_user1_shell:\n"

    out = check_output(['sss_override', 'user-show',
                        'user2@LDAP']).decode('utf-8')
    assert out == "user2@LDAP:ov_user2:10020:20020:Overriden User 2:"\
                  "/home/ov/user2:/bin/ov_user2_shell:\n"

    # Return error on non-existing user
    ret = subprocess.call(['sss_override', 'user-show', 'nonexisting_user'])
    assert ret == 1


#
# Override user-find
#


@pytest.fixture
def env_find_user_override(request, ldap_conn,
                           env_two_users_and_group_overriden):
    pass


def test_find_user_override(ldap_conn, env_find_user_override):

    out = check_output(['sss_override', 'user-find']).decode('utf-8')

    # Expected override of users
    exp_usr_ovrd = ['user1@LDAP:ov_user1:10010:20010:Overriden User 1:'
                    '/home/ov/user1:/bin/ov_user1_shell:',
                    'user2@LDAP:ov_user2:10020:20020:Overriden User 2:'
                    '/home/ov/user2:/bin/ov_user2_shell:']

    assert set(out.splitlines()) == set(exp_usr_ovrd)

    out = check_output(['sss_override', 'user-find', '--domain=LDAP'])

    assert set(out.decode('utf-8').splitlines()) == set(exp_usr_ovrd)

    # Unexpected parameter is reported
    ret = subprocess.call(['sss_override', 'user-find', 'PARAM'])
    assert ret == 1


#
# Group tests
#


#
# Common group asserts
#

def assert_group_overriden(override_name=True):

    # Assert entries are overridden
    empty_group = dict(gid=3002, mem=ent.contains_only())
    group = dict(gid=3001, mem=ent.contains_only("user1", "user2"))

    ent.assert_group_by_name("group", group)
    ent.assert_group_by_name("group@LDAP", group)

    if override_name:
        ent.assert_group_by_name("ov_group", group)
        ent.assert_group_by_name("ov_group@LDAP", group)

    ent.assert_group_by_name("empty_group", empty_group)
    ent.assert_group_by_name("empty_group@LDAP", empty_group)

    if override_name:
        ent.assert_group_by_name("ov_empty_group", empty_group)
        ent.assert_group_by_name("ov_empty_group@LDAP", empty_group)


def assert_group_default():

    # Assert entries are not overridden
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_group')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_group@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_empty_group')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_empty_group@LDAP')

    empty_group = dict(gid=2002, mem=ent.contains_only())
    group = dict(gid=2001, mem=ent.contains_only("user1", "user2"))

    ent.assert_group_by_name("group", group)
    ent.assert_group_by_name("group@LDAP", group)
    ent.assert_group_by_name("empty_group", empty_group)
    ent.assert_group_by_name("empty_group@LDAP", empty_group)


#
# Common fixtures for groups
#


@pytest.fixture
def env_group_basic(request, ldap_conn):
    prepare_sssd(request, ldap_conn)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001,
                      gecos='User Number 1',
                      loginShell='/bin/user1_shell',
                      homeDirectory='/home/user1')

    ent_list.add_user("user2", 10002, 20001,
                      gecos='User Number 2',
                      loginShell='/bin/user2_shell',
                      homeDirectory='/home/user2')

    ent_list.add_group("group", 2001,
                       ["user2", "user1"])
    ent_list.add_group("empty_group", 2002, [])

    create_ldap_fixture(request, ldap_conn, ent_list)

    # Assert entries are not overriden
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_group')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_group@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_empty_group')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_empty_group@LDAP')


@pytest.fixture
def env_group_override(request, ldap_conn, env_group_basic):

    # Override
    subprocess.check_call(["sss_override", "group-add", "group",
                           "-n", "ov_group",
                           "-g", "3001"])

    subprocess.check_call(["sss_override", "group-add", "empty_group@LDAP",
                           "--name", "ov_empty_group",
                           "--gid", "3002"])

    # Restart SSSD so the override might take effect
    restart_sssd()

    # Assert entries are overridden
    assert_group_overriden()


#
# Simple group override
#


@pytest.fixture
def env_simple_group_override(request, ldap_conn, env_group_basic):

    # Override
    subprocess.check_call(["sss_override", "group-add", "group",
                           "-n", "ov_group",
                           "-g", "3001"])

    subprocess.check_call(["sss_override", "group-add", "empty_group@LDAP",
                           "--name", "ov_empty_group",
                           "--gid", "3002"])

    # Restart SSSD so the override might take effect
    restart_sssd()


def test_simple_group_override(ldap_conn, env_simple_group_override):
    """Test entries are overriden"""

    assert_group_overriden()


#
# Root group override
#


@pytest.fixture
def env_root_group_override(request, ldap_conn, env_group_basic):

    # Override
    subprocess.check_call(["sss_override", "group-add", "group",
                           "-n", "ov_group",
                           "-g", "0"])

    subprocess.check_call(["sss_override", "group-add", "empty_group@LDAP",
                           "--name", "ov_empty_group",
                           "--gid", "0"])

    # Restart SSSD so the override might take effect
    restart_sssd()


def test_root_group_override(ldap_conn, env_root_group_override):
    """Test entries are overriden"""

    group = dict(gid=2001, mem=ent.contains_only("user1", "user2"))
    empty_group = dict(gid=2002, mem=ent.contains_only())

    ent.assert_group_by_name("group", group)
    ent.assert_group_by_name("ov_group", group)
    ent.assert_group_by_name("group@LDAP", group)
    ent.assert_group_by_name("ov_group@LDAP", group)
    ent.assert_group_by_name("empty_group", empty_group)
    ent.assert_group_by_name("ov_empty_group", empty_group)
    ent.assert_group_by_name("empty_group@LDAP", empty_group)
    ent.assert_group_by_name("ov_empty_group@LDAP", empty_group)


#
# Replace group override
#


@pytest.fixture
def env_replace_group_override(request, ldap_conn, env_group_override):

    # Override of override
    subprocess.check_call(["sss_override", "group-add", "group",
                           "-n", "ov2_group",
                           "-g", "4001"])

    subprocess.check_call(["sss_override", "group-add", "empty_group@LDAP",
                           "--name", "ov2_empty_group",
                           "--gid", "4002"])

    # Restart SSSD so the override might take effect
    restart_sssd()


def test_replace_group_override(ldap_conn, env_replace_group_override):

    # Test overrides are overridden
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_group')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_group@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_empty_group')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_empty_group@LDAP')

    group = dict(gid=4001, mem=ent.contains_only("user1", "user2"))
    empty_group = dict(gid=4002, mem=ent.contains_only())

    ent.assert_group_by_name("group", group)
    ent.assert_group_by_name("ov2_group", group)
    ent.assert_group_by_name("group@LDAP", group)
    ent.assert_group_by_name("ov2_group@LDAP", group)

    ent.assert_group_by_name("empty_group", empty_group)
    ent.assert_group_by_name("empty_group@LDAP", empty_group)
    ent.assert_group_by_name("ov2_empty_group", empty_group)
    ent.assert_group_by_name("ov2_empty_group@LDAP", empty_group)


#
# Remove group override
#


@pytest.fixture
def env_remove_group_override(request, ldap_conn, env_group_override):

    # Drop all overrides
    subprocess.check_call(["sss_override", "group-del", "group"])
    subprocess.check_call(["sss_override", "group-del", "empty_group@LDAP"])

    # Avoid hitting memory cache
    time.sleep(2)


def test_remove_group_override(ldap_conn, env_remove_group_override):

    # Test overrides were dropped
    assert_group_default()


#
# Override group import/export
#


@pytest.fixture
def env_imp_exp_group_override(request, ldap_conn, env_group_override):

    # Export overrides
    subprocess.check_call(["sss_override", "group-export",
                           OVERRIDE_FILENAME])

    # Drop all overrides
    subprocess.check_call(["sss_override", "group-del", "group"])
    subprocess.check_call(["sss_override", "group-del", "empty_group@LDAP"])

    # Avoid hitting memory cache
    time.sleep(2)

    assert_group_default()

    # Import overrides
    subprocess.check_call(["sss_override", "group-import",
                           OVERRIDE_FILENAME])
    restart_sssd()


def test_imp_exp_group_override(ldap_conn, env_imp_exp_group_override):

    assert_group_overriden()


# Regression test for bug 3179


def test_imp_exp_group_override_noname(ldap_conn, env_group_basic):

    # Override - do not use -n here)
    subprocess.check_call(["sss_override", "group-add", "group",
                           "-g", "3001"])

    subprocess.check_call(["sss_override", "group-add", "empty_group@LDAP",
                           "--gid", "3002"])

    # Restart SSSD so the override might take effect
    restart_sssd()

    # Assert entries are overridden
    assert_group_overriden(override_name=False)

    # Export overrides
    subprocess.check_call(["sss_override", "group-export",
                           OVERRIDE_FILENAME])

    # Drop all overrides
    subprocess.check_call(["sss_override", "group-del", "group"])
    subprocess.check_call(["sss_override", "group-del", "empty_group@LDAP"])

    # Avoid hitting memory cache
    time.sleep(2)

    assert_group_default()

    # Import overrides
    subprocess.check_call(["sss_override", "group-import",
                           OVERRIDE_FILENAME])
    restart_sssd()

    assert_group_overriden(override_name=False)


# Regression test for bug #2802
# sss_override segfaults when accidentally adding --help flag to some commands


@pytest.fixture
def env_regr_2802_override(request, ldap_conn):

    prepare_sssd(request, ldap_conn)


def test_regr_2802_override(ldap_conn, env_regr_2802_override):

    subprocess.check_call(["sss_override", "user-del", "--help"])


# Regression test for bug #2757
# sss_override does not work correctly when 'use_fully_qualified_names = True'


@pytest.fixture
def env_regr_2757_override(request, ldap_conn):

    prepare_sssd(request, ldap_conn, use_fully_qualified_names=True)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001)

    create_ldap_fixture(request, ldap_conn, ent_list)

    # Assert entries are not overridden
    ent.assert_passwd_by_name(
        'user1@LDAP',
        dict(name='user1@LDAP', passwd='*', uid=10001, gid=20001))
    with pytest.raises(KeyError):
        pwd.getpwnam('alias1')
    with pytest.raises(KeyError):
        pwd.getpwnam('alias1@LDAP')

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1@LDAP",
                           "-n", "alias1"])
    restart_sssd()


def test_regr_2757_override(ldap_conn, env_regr_2757_override):

    # Assert entries are overridden
    ent.assert_passwd_by_name(
        'user1@LDAP',
        dict(name='alias1@LDAP', passwd='*', uid=10001, gid=20001))
    ent.assert_passwd_by_name(
        'alias1@LDAP',
        dict(name='alias1@LDAP', passwd='*', uid=10001, gid=20001))

    with pytest.raises(KeyError):
        pwd.getpwnam('user1')
    with pytest.raises(KeyError):
        pwd.getpwnam('alias1')


# Regression test for bug #2790
# sss_override --name doesn't work with RFC2307 and ghost users


@pytest.fixture
def env_regr_2790_override(request, ldap_conn):

    prepare_sssd(request, ldap_conn)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001)
    ent_list.add_user("user2", 10002, 20002)
    ent_list.add_group("group1", 2001,
                       ["user1", "user2"])
    ent_list.add_group("group2", 2002,
                       ["user2"])

    create_ldap_fixture(request, ldap_conn, ent_list)

    # Assert entries are not overridden
    with pytest.raises(KeyError):
        pwd.getpwnam('alias1')
    with pytest.raises(KeyError):
        pwd.getpwnam('alias1@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('alias2')
    with pytest.raises(KeyError):
        pwd.getpwnam('alias2@LDAP')

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1",
                           "-n", "alias1"])
    subprocess.check_call(["sss_override", "user-add", "user2",
                           "-n", "alias2"])

    restart_sssd()


def test_regr_2790_override(ldap_conn, env_regr_2790_override):

    # Assert entries are overridden
    (res, errno, grp_list) = sssd_id.get_user_groups("alias1")
    assert res == sssd_id.NssReturnCode.SUCCESS, \
        "Could not find groups for user1 %d" % errno
    assert sorted(grp_list) == sorted(["20001", "group1"])

    (res, errno, grp_list) = sssd_id.get_user_groups("alias2")
    assert res == sssd_id.NssReturnCode.SUCCESS, \
        "Could not find groups for user2 %d" % errno
    assert sorted(grp_list) == sorted(["20002", "group1", "group2"])


# Test fully qualified and case-insensitive names
@pytest.fixture
def env_mix_cased_name_override(request, ldap_conn):
    """Setup test for mixed case names"""

    prepare_sssd(request, ldap_conn, True, False)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001)
    ent_list.add_user("uSeR2", 10002, 20002)

    create_ldap_fixture(request, ldap_conn, ent_list)

    pwd.getpwnam('user1@LDAP')
    pwd.getpwnam('user2@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user1@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user2@LDAP')

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1@LDAP",
                           "-u", "10010",
                           "-g", "20010",
                           "-n", "ov_user1",
                           "-c", "Overriden User 1",
                           "-h", "/home/ov/user1",
                           "-s", "/bin/ov_user1_shell"])

    subprocess.check_call(["sss_override", "user-add", "user2@LDAP",
                           "-u", "10020",
                           "-g", "20020",
                           "-n", "ov_user2",
                           "-c", "Overriden User 2",
                           "-h", "/home/ov/user2",
                           "-s", "/bin/ov_user2_shell"])

    restart_sssd()


def test_mix_cased_name_override(ldap_conn, env_mix_cased_name_override):
    """Test if names with upper and lower case letter are overridden"""

    # Assert entries are overridden
    user1 = dict(name='ov_user1@LDAP', passwd='*', uid=10010, gid=20010,
                 gecos='Overriden User 1',
                 dir='/home/ov/user1',
                 shell='/bin/ov_user1_shell')

    user2 = dict(name='ov_user2@LDAP', passwd='*', uid=10020, gid=20020,
                 gecos='Overriden User 2',
                 dir='/home/ov/user2',
                 shell='/bin/ov_user2_shell')

    ent.assert_passwd_by_name('user1@LDAP', user1)
    ent.assert_passwd_by_name('ov_user1@LDAP', user1)

    ent.assert_passwd_by_name('user2@LDAP', user2)
    ent.assert_passwd_by_name('ov_user2@LDAP', user2)


# Test with override_homedir option
@pytest.fixture
def env_override_homedir_option(request, ldap_conn):
    """Setup test for override_homedir option and overrides"""

    prepare_sssd(request, ldap_conn, override_homedir_option=True)

    # Add entries
    ent_list = ldap_ent.List(ldap_conn.ds_inst.base_dn)
    ent_list.add_user("user1", 10001, 20001)
    ent_list.add_user("user2", 10002, 20002)

    create_ldap_fixture(request, ldap_conn, ent_list)

    pwd.getpwnam('user1@LDAP')
    pwd.getpwnam('user2@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user1@LDAP')
    with pytest.raises(KeyError):
        pwd.getpwnam('ov_user2@LDAP')

    # Override
    subprocess.check_call(["sss_override", "user-add", "user1@LDAP",
                           "-u", "10010",
                           "-g", "20010",
                           "-n", "ov_user1",
                           "-c", "Overriden User 1",
                           # no homedir override
                           "-s", "/bin/ov_user1_shell"])

    subprocess.check_call(["sss_override", "user-add", "user2@LDAP",
                           "-u", "10020",
                           "-g", "20020",
                           "-n", "ov_user2",
                           "-c", "Overriden User 2",
                           "-h", "/home/ov/user2",
                           "-s", "/bin/ov_user2_shell"])

    restart_sssd()


def test_override_homedir_option(ldap_conn, env_override_homedir_option):
    """Test if overrides will overwrite override_homedir option"""

    # Assert entries are overridden, user1 has no homedir override and
    # override_homedir option should be used, user2 has a homedir override
    # which should be used.
    user1 = dict(name='ov_user1', passwd='*', uid=10010, gid=20010,
                 gecos='Overriden User 1',
                 dir='/home/ov_option/ov_user1',
                 shell='/bin/ov_user1_shell')

    user2 = dict(name='ov_user2', passwd='*', uid=10020, gid=20020,
                 gecos='Overriden User 2',
                 dir='/home/ov/user2',
                 shell='/bin/ov_user2_shell')

    ent.assert_passwd_by_name('user1@LDAP', user1)
    ent.assert_passwd_by_name('ov_user1@LDAP', user1)
    ent.assert_passwd_by_name('user1', user1)
    ent.assert_passwd_by_name('ov_user1', user1)

    ent.assert_passwd_by_name('user2@LDAP', user2)
    ent.assert_passwd_by_name('ov_user2@LDAP', user2)
    ent.assert_passwd_by_name('user2', user2)
    ent.assert_passwd_by_name('ov_user2', user2)
